# -*- coding: utf-8 -*-
"""Management of all migrations.

Handles actions like upgrade, downgrading, keeping track of
which migrations did run and which not. Used by :py:mod:`.MigrationCommand`.
"""
from __future__ import print_function
import os
import glob
import imp
import inflection
import datetime
from tabulate import tabulate
from termcolor import colored
from experimentum.cli import print_failure


class Migrator(object):

    """Management of all migrations.

    Attributes:
        path (str): Path to the migrations folder.
    """

    @staticmethod
    def get_migration_files(path):
        """Get all of the migration files in a given path.

        Args:
            path (str): Path to migrations folder

        Returns:
            list: list of migration files
        """
        files = glob.glob(os.path.join(path, '[0-9]*_*.py'))

        if not files:
            return []

        files = list(map(lambda f: os.path.basename(f).replace('.py', ''), files))
        files = sorted(files)

        return files

    def __init__(self, path, app):
        """Set the path for the migrations.

        Args:
            path (str): Path to the migrations folder.
        """
        self.path = path

        # create .version file
        name = '{}/.version'.format(self.path)

        if not os.path.isfile(name):
            with open(name, 'w+') as file:
                file.write('')

        # load migrations
        self.migrations = {}
        names = self.get_migration_files(path)

        for name in names:
            file = os.path.join(path, '%s.py' % name)
            with open(file, 'rb') as filehandler:
                mod = imp.load_source('migrations', file, filehandler)
                migration = getattr(mod, inflection.camelize(name[15:]))
                self.migrations[name] = migration(app)

    def status(self, printing=True):
        """Print the current status of which migrations did run and which not.

        Args:
            printing (bool, optional): Defaults to True. Whether or not the status table is printed.

        Returns:
            list: migration status list
        """
        # check whether migration did run or not
        migrated = self._get_migrated_revisions()
        migrations = self.get_migration_files(self.path)
        status = [
            [colored(m, 'cyan'), colored('Yes', 'green')] if m[:14] in migrated
            else [colored(m, 'cyan'), colored('No', 'red')]
            for m in migrations
        ]

        headers = [colored('Migration', 'yellow'), colored('Ran?', 'yellow')]

        # Print Table
        if printing:
            print(tabulate(status, headers=headers, tablefmt='psql'))

        return status

    def make(self, name):
        """Make a new migration file.

        Args:
            name (str): Name for the migration.
        """
        name = ''.join(list(filter(lambda x: x.isalpha() or x == '_', name.replace(' ', '_'))))
        migration = inflection.camelize(name)
        revision = datetime.datetime.now().strftime('%Y%m%d%H%M%S')
        filename = "{}/{}_{}.py".format(self.path, revision, name.lower())

        stub = ''
        with open(os.path.join(os.path.dirname(__file__), 'migration.stub'), 'r') as content:
            stub = content.read()

        with open(filename, 'w+') as file:
            file.write(stub.format(migration=migration, name=name, revision=revision))
            print(colored('› Migration created successfully!', 'green'))

    def up(self, migration=None):
        """Upgrade to a new migration revision.

        Args:
            migration (Migration, optional): Defaults to None. The Migration Class to upgrade to.

        Raises:
            TypeError: if migration is not a valid migration
        """
        # Select first migration to run
        if migration is None:
            migrated = self._get_migrated_revisions()
            migrations = self.get_migration_files(self.path)
            migration = [mig for mig in migrations if mig[:14] not in migrated]

            if len(migration) == 0:
                print(colored('› Migrations are all up to date.', 'green'))
                return

            migration = self.migrations.get(migration[0])

        # update migration
        try:
            migration.up()
            self._update_revision(migration.revision)
            print(colored('› Migrated', 'green'), colored(migration, 'cyan'))
        except Exception as exc:
            print_failure('Error while ugrading migration {}: {}'.format(migration, exc), 1)

    def down(self, migration=None):
        """Downgrade to an old migration revision.

        Args:
            migration (Migration, optional): Defaults to None. The Migration Class to upgrade to.

        Raises:
            TypeError: if migration is not a valid migration
        """
        # Select first migration to run
        if migration is None:
            # get finished migrations
            migrated = self._get_migrated_revisions()
            migrations = self.get_migration_files(self.path)[::-1]
            migration = [mig for mig in migrations if mig[:14] in migrated]

            if len(migration) == 0:
                print(colored('× There are no migrations to downgrade.', 'red'))
                return

            migration = self.migrations.get(migration[0])

        # downgrade migration
        try:
            migration.down()
            self._update_revision(migration.revision, delete=True)
            print(colored('› Migrated', 'green'), colored(migration, 'cyan'))
        except Exception as exc:
            print_failure('Error while downgrading migration {}: {}'.format(migration, exc), 1)

    def refresh(self):
        """Rerun all migrations."""
        migrated = self._get_migrated_revisions()
        migrations = self.get_migration_files(self.path)

        # downgrade all migrations
        print(colored('--- Downgrading Migrations ---', 'yellow'))
        for migration in migrations[::-1]:
            if migration[:14] in migrated:
                self.down(self.migrations.get(migration))

        # run all up migrations
        print(colored('\n--- Upgrading Migrations ---', 'yellow'))
        for migration in migrations:
            self.up(self.migrations.get(migration))

        return True

    def _update_revision(self, revision, delete=False):
        """Update the revision number.

        Args:
            revision (str): The new revision
            delete (bool, optional): Defaults to False. Whether to delete the revision or not.
        """
        fname = os.path.join(self.path, '.version')
        revisions = self._get_migrated_revisions()

        with open(fname, 'w+') as filehandler:
            if delete:
                # print('remove revision')
                revisions.remove(revision)
            else:
                # print('add revision')
                revisions.append(revision)

            filehandler.write('|'.join(revisions))

    def _get_migrated_revisions(self):
        """Get old migrations revisions.

        Returns:
            list
        """
        old = []
        with open('{}/.version'.format(self.path), 'r') as filehandler:
            old = filehandler.read().split('|')
        return old



"""Factory to load and init plot classes."""
from experimentum.cli import print_failure
from experimentum.utils import find_files, load_class
from experimentum.Plots import AbstractPlot


class Factory(object):

    """Create instances of plot classes based on their name.

    Attributes:
        app (App): Main Service Container.
    """

    def __init__(self, app):
        """Init Plot factory.

        Args:
            app (App): Main Service Container.
        """
        self.app = app

    def create(self, name):
        """Create a new plot class instance.

        Args:
            name (str): Name of the plot.

        Returns:
            AbstractPlot: Plot Class
        """
        # Get plot data from config and validate it
        data = self.parse(name)
        config = self.validate(data)

        # Load plot class
        path = self.app.config.get('app.plots.path', 'plots')
        plots = find_files(self.app.root, path, name, '(plot|chart|graph).py')

        if not plots:
            print_failure(
                'Could not find plot named "{}" under path "{}"'.format(name, path),
                exit_code=1
            )
        plot = load_class(plots[0], path, AbstractPlot)

        # Init plot class
        return plot(
            self.app.repositories.get('ExperimentRepository'),
            config,
            config.get('type', 'plot')
        )

    def parse(self, plot):
        """Parse the config file for the current plot.

        Args:
            plot (str): Name of the plot.

        Returns:
            object: Config data for the plot
        """
        config = self.app.config.get('plots.{}'.format(plot), None)

        if config is None:
            print_failure(
                'Could not parse config for plot "%s". '
                'Make sure it exists in the config file!' % plot,
                exit_code=1
            )

        return config

    def validate(self, data):
        """Validate the config data and save it in the Config Container if valid.

        Args:
            data (object): Config Data

        Returns:
            Config: Config Data for the plot.
        """
        # TODO: Actually validate config data before saving
        config = self.app.make('config')
        config.set(data)

        return config



"""Provides an easy to use web interface via flask.

Uses the Application Factory Pattern to create a new flask app instance.
"""
from flask import Flask, render_template
from werkzeug.serving import run_simple
import os


class Server(object):

    """Simple wrapper for the webserver."""

    def __init__(self, app, testing=False):
        """Init the server.

        Args:
            app (App): Main Service Container.
            testing (bool, optional): Defaults to False. Enable/Disable test mode.
        """
        self.app = app
        self.testing = testing

    def create_app(self):
        """Create a new flask application.

        Returns:
            Flask: Flask App
        """
        if self.testing:
            return create_app(self.app, {'TESTING': True})

        return create_app(self.app)

    def run(self, port=5000, debugger=False, reloader=True):
        """Run the server.

        Args:
            port (int, optional): Defaults to 5000. Port to run on.
            debugger (bool, optional): Defaults to False. Enable/Disable Debugger.
            reloader (bool, optional): Defaults to True. Enable/Disable Reloader.
        """
        run_simple(
            'localhost', port, self.create_app(), use_reloader=reloader,
            use_debugger=debugger, use_evalex=True, threaded=True
        )


def create_app(container, test_config=None):
    """Create a new flask app.

    Uses the Application Factory Pattern: http://flask.pocoo.org/docs/1.0/patterns/appfactories/

    Args:
        container (App): Main Service Container
        test_config (dict, optional): Defaults to None. Test Configuration

    Returns:
        Flask: Flask App
    """
    # Create Flask App
    app = Flask(__name__, instance_relative_config=True)
    app.jinja_env.auto_reload = True
    # app.template_folder = os.path.abspath(os.path.join(
    #     os.path.dirname(__file__),
    #     'templates'
    # ))
    # print(app.template_folder)
    app.config.from_mapping(
        # Auto-reload templates on change
        TEMPLATES_AUTO_RELOAD=True,

        # Folder where plots are stored
        UPLOAD_FOLDER=os.path.abspath(os.path.join(container.root, '.tmp'))
    )
    app.config['container'] = container

    # load the instance config, if it exists, when not testing, or the test_config
    if test_config is None:
        app.config.update(container.config.get('app.webgui', {}))
    else:
        app.config.update(test_config)

    # 404 Page
    @app.errorhandler(404)
    def page_not_found(err):
        """Return 404 error page.

        Args:
            err (werkzeug.exceptions.NotFound): Error message

        Returns:
            Response: 404 Page
        """
        return render_template('404.jinja', msg=err), 404

    # Add Blueprints
    from .views import dashboard, migrations, experiments, plots
    app.register_blueprint(dashboard.blueprint)
    app.register_blueprint(migrations.blueprint, url_prefix='/migrations')
    app.register_blueprint(experiments.blueprint, url_prefix='/experiments')
    app.register_blueprint(plots.blueprint, url_prefix='/plots')

    return app



# -*- coding: utf-8 -*-
"""Run experiments and save the results for future analysations.

Writing Experiments
-------------------
Experiments are created in the `experiments` directory and they must adhere to
the following naming convention: `{NAME}Experiment.py`.

All Experiments extend the :py:class:`.Experiment` class. Experiments
contain a :py:meth:`~.Experiment.reset` and :py:meth:`~.Experiment.run`
method. Within :py:meth:`~.Experiment.reset` the method, you should
reset/initialize the data structuresand values you want to use in each test run.

The :py:meth:`~.Experiment.run` method should contain the code you want to test.
It should return a dictionary with the values you want to save.

The Reset Method
----------------
As mentioned before the :py:meth:`~.Experiment.reset` should reset/initialize the
data structuresand values you want to use in each test run.

Let's take a look at a basic experiment. Within any of your experiment methods,
you always have access to the :py:attr:`~.Experiment.app` attribute which
provides access to the main app class and to the :py:attr:`~.Experiment.config` which
contains the content of the :py:attr:`~.Experiment.config_file`::

    from experimentum.Experiments import Experiment
    import random


    class FooExperiment(Experiment):

        config_file = 'foo.json'

        def reset(self):
            # use app to create an instance of a custom aliased class
            self.user = self.config.get('user')
            self.some_class = self.app.make('some_class', self.user)
            self.rand = random.randint(0, 10)


The Run Method
--------------
As mentioned before the :py:meth:`~.Experiment.run` method should contain the code
you want to test and return a dictionary with the values you want to save.

Let's take a look at a basic experiment, assuming that you added a ``rand`` attribute to
your TestCaseRepository with a migration::

    from experimentum.Experiments import Experiment
    import random


    class FooExperiment(Experiment):

        config_file = 'foo.json'

        def run(self):
            with self.performance.point('Task to measure some Rscript algo') as point:
                script = self.call('some_script.r')  # prints json to stdout as return value.
                algo_result = script.get_json()
                script.process.wait()  # Wait for child process to terminate.

            # Add a custom performance entry
            return {
                'rand': self.rand,
                'performances': [{
                    'label': 'Custom Rscript measuring',
                    'time': algo_result.get('time'),
                    'memory': 0,
                    'peak_memory': 0,
                    'level': 0,
                    'type': 'custom'
                }]
            }
"""
from __future__ import print_function
import os
import glob
import subprocess
import json
from datetime import datetime
from six import add_metaclass
from abc import abstractmethod, ABCMeta
from experimentum.Config import Config
from experimentum.Experiments import Performance
from experimentum.cli import print_progress, print_failure
from experimentum.utils import get_basenames, load_class, find_files


class Script(object):

    """Call another script to run algorithms for your experiment.

    Example::

        script = Script(['Rscript', 'myalgo.r', 'arg1', 'arg2'], verbose=True, shell=True)
        algo_result = script.get_json()
        script.process.wait()  # Wait for child process to terminate.

    Attributes:
        process (subprocess.Popen): Called script.
        output (str): Output of called script.
    """

    def __init__(self, cmd, verbose=False, shell=False, stdout=subprocess.PIPE):
        """Get the return vaue of a process (i.e, the last print statement).

        .. Warning::
            Passing ``shell=True`` can be a security hazard if combined with untrusted input.
            See the warning under `Frequently Used Arguments
            <https://docs.python.org/2/library/subprocess.html#frequently-used-arguments>`_
            for details.

        Args:
            cmd (str, list): Command which you want to call.
            verbose (bool, optional): Defaults to False. Print the cmd output or not.
            shell (bool, optional): Defaults to False. Specifices whether to use the
                shell as the program to execute.
            stdout (int, optional): Defaults to subprocess.PIPE. Specify standard output.
        """
        self.process = subprocess.Popen(cmd, stdout=stdout, shell=shell)
        self.output = None

        # poll will return the exit code if the process is completed otherwise it returns null
        while self.process.poll() is None:
            line = self.process.stdout.readline()
            if not line:
                break
            self.output = line  # last print statement
            if verbose:
                print(line.rstrip().decode('utf-8'))

    def get_json(self):
        """Decode JSON of process output.

        Returns:
            object: Process output
        """
        return json.loads(self.get_text())

    def get_text(self):
        """Get the text of the process output.

        Returns:
            str: Process output
        """
        return self.output


@add_metaclass(ABCMeta)
class Experiment(object):

    """Run experiments and save the results in the data store.

    Attributes:
        app (App): Main Application Class.
        performance (Performance): Performance Profiler.
        config (Config): Hold the experiment configuration.
        show_progress (bool): Flag to show/hide the progress bar.
        hide_performance (bool): Flag to show/hide the performance table.
        config_file (str): Config file to load.
        repos (dict): Experiment and Testcast Repo to save results.
    """
    config_file = None

    def __init__(self, app, path):
        """Init the experiment.

        Args:
            app (App): Main Application class
            path (str): Path to experiments folder
        """
        self.app = app
        self.performance = Performance()
        self.config = Config()
        self.show_progress = False
        self.hide_performance = False
        self.repos = {'experiment': None, 'testcase': None}
        self._path = path

    @staticmethod
    def get_experiments(path):
        """[DEPRECATED] Get experiment names from exp files/classes.

        Args:
            path (str): Path to experiments folder.

        Returns:
            list: Names of experiments
        """
        # TODO: Deprecated remove!
        print('[DEPRECATED]: Remove Experiments.get_experiments usage!!')
        files = glob.glob(os.path.join(path, '[!_]*.py'))
        return list(map(
            lambda exp: os.path.basename(exp).lower().replace('experiment.py', ''),
            files
        ))

    @staticmethod
    def get_status(app):
        """Get status information about experiments.

        Args:
            app (App): Main Service Provider/Container.

        Returns:
            dict: Dictionary with experiment status
        """
        # Load experiment classes
        path = app.config.get('app.experiments.path', 'experiments')
        exps = get_basenames(app.root, path, 'experiment.py')
        data = {exp.lower(): {'count': 0, 'name': exp} for exp in exps}

        # Load experiment stats
        repo = app.repositories.get('ExperimentRepository')
        rows = repo.all()
        for exp in rows:
            idx = exp.name.lower()

            # Exp file does not exist anymore
            if idx not in data:
                data[idx] = {'count': 0, 'name': exp.name, 'missing': True}

            data[idx]['count'] += 1

            if exp.config_file:
                data[idx]['config_file'] = exp.config_file

        return data

    @staticmethod
    def load(app, path, name):
        """Load and initialize an experiment class.

        Args:
            app (App): Main app calss
            path (str): Path to experiments folder.
            name (str): Name of experiment.

        Returns:
            Experiment: Loaded experiment.
        """
        # Find Experiment Files
        files = find_files(app.root, path, name, remove='experiment.py')
        if not files:
            print_failure(
                'Could not find experiment named "{}" under path "{}"'.format(name, path),
                exit_code=1
            )

        # Load Experiment class if possible
        experiment = load_class(files[0], 'experiments', Experiment)
        return experiment(app, path)

    @staticmethod
    def call(cmd, verbose=False, shell=False):
        """Call another script to run algorithms for your experiment.

        .. Warning::
            Passing ``shell=True`` can be a security hazard if combined with untrusted input.
            See the warning under `Frequently Used Arguments
            <https://docs.python.org/2/library/subprocess.html#frequently-used-arguments>`_
            for details.

        Args:
            cmd (str, list): Command which you want to call.
            verbose (bool, optional): Defaults to False. Print the cmd output or not.
            shell (bool, optional): Defaults to False. Specifices whether to use the
                shell as the program to execute.

        Returns:
            Script: Executed script to get output from
        """
        return Script(cmd, verbose, shell)

    def boot(self):
        """Boot up the experiment, e.g. load config etc."""
        # Load Config/Args for experiment
        if self.config_file:
            try:
                with open(os.path.join(self._path, self.config_file), 'r') as cfg:
                    self.config.set(json.load(cfg))
            except Exception as exc:
                print_failure(exc, 2)

        # Load Experiment and testcase repos
        try:
            self.repos['experiment'] = self.app.repositories.get('ExperimentRepository')
            self.repos['testcase'] = self.app.repositories.get('TestCaseRepository')

            self.repos['experiment'] = self.repos['experiment'].from_dict({
                'name': self.__class__.__name__.replace('Experiment', ''),
                'start': datetime.now(),
                'config_file': self.config_file,
                'config_content': json.dumps(self.config.all()),
                'tests': []
            })
            self.repos['experiment'].create()
        except Exception as exc:
            print_failure(exc, 2)

    def start(self, steps=10):
        """Start the test runs of the experiment.

        Args:
            steps (int, optional): Defaults to 10. How many tests runs should be executed.
        """
        # Booting
        with self.performance.point('Booting Experiment'):
            self.boot()

        # Running tests
        for iteration in self.performance.iterate(1, steps):
            # Reset test state
            result = None
            self.reset()

            # Run experiment
            with self.performance.point('Runing Experiment'):
                result = self.run()

            # Save Results
            if result:
                self.save(result, iteration)
            else:
                msg = 'Experiment returned an empty result. Are you sure this is correct?'
                self.app.log.warning(msg)
                print('[WARNING]: ' + msg)

            if self.show_progress:
                print_progress(iteration, steps, prefix='Progress:', suffix='Complete')

        # Finished Experiment
        self.repos['experiment'].finished = datetime.now()
        self.repos['experiment'].update()
        if self.hide_performance is False:
            self.performance.results()

    def save(self, result, iteration):
        """Save the test results in the data store.

        Args:
            result (dict): Result of experiment test run.
            iteration (int): Number of test run iteration.
        """
        data = {
            'experiment_id':  self.repos['experiment'].id,
            'iteration': iteration,
            'performances': []
        }
        data.update(result)
        data['performances'].extend(self.performance.export())

        try:
            self.repos['testcase'].from_dict(data).create()
        except Exception as exc:
            for msg in str(exc).split('\n'):
                print_failure(msg)
            raise SystemExit(-1)

    @abstractmethod
    def reset(self):
        """Reset data structured and values used in the run method."""
        raise NotImplementedError('Must implement reset method.')

    @abstractmethod
    def run(self):
        """Run a test of the experiment."""
        raise NotImplementedError('Must implement run method.')



"""Import classes for easier importing by other packages/modules."""
# flake8: noqa
from .Experiments import App, Performance, Experiment
from .Config import Config, Loader
from .Commands import AbstractCommand, CommandManager, MigrationCommand
from .Storage import AbstractStore, Blueprint, Column, ForeignKey, Migration, Migrator,\
    Schema, ColumnFactory, Platform, SQLitePlatform, Store

from colorama import init

# use Colorama to make Termcolor work on Windows too
init()

##################
# Helper Methods #
##################
# def config(get=None, default=None):
#     """Shortcut for getting the config instance."""
#     return app().config.get(get, default) if get else app().config



"""Adding CLI commands to the app and handle their execution."""
import argparse
import types
import sys
from termcolor import colored
from experimentum.cli import print_failure
from experimentum.Commands import AbstractCommand


class _Parser(argparse.ArgumentParser):

    """Helper to color in titles."""

    def titles(self, positionals, optionals, color='cyan'):
        """Color in titles.

        Args:
            positionals (str): Positionals Title
            optionals (str): Optionals Title
            color (str, optional): Defaults to 'cyan'. Text Color
        """
        self._positionals.title = colored(positionals, color)
        self._optionals.title = colored(optionals, color)


class CommandManager(object):

    """CommandManager class registers and manages commands.

    Attributes:
        app (App): Main App class.
        commands (dict): Registered commands.
    """
    commands = {}

    def __init__(self, app, prog, description=''):
        """Create and setup up the argument parser and superparsers.

        Args:
            app (App): Main App class
            prog (str): Name of the programm.
            description (str, optional): Defaults to ''. Description of the programm.
        """
        self.app = app

        # Init ArgumentParser
        self._parser = _Parser(
            add_help=False,
            prog=colored(prog, 'yellow'),
            description=colored(description, 'yellow'),
            formatter_class=ColoredHelpFormatter
        )
        self._parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show this help message and exit.'
        )
        self._parser.add_argument(
            '-v', '--version',
            action='version',
            version='%(prog)s 1.0',
            help="Show program's version number and exit."
        )

        # Colore Titles
        self._parser.titles('Arguments', 'Options', color='cyan')

        # Init subparsers
        self._subparsers = self._parser.add_subparsers()

    def add_command(self, name, cmd):
        """Add a new command to the parser.

        Args:
            name (str): Name of the command
            cmd (function, AbstractCommand): Command Handler
        """
        if (
            not isinstance(cmd, types.FunctionType) and
            not issubclass(cmd, AbstractCommand)
        ):
            print_failure("{}-Command must inherit from AbstractCommand!".format(name), 1)

        # setup command
        cmd = cmd()  # type: AbstractCommand
        command = self._subparsers.add_parser(
            name,
            help=cmd.help,
            description=colored(cmd.description, 'yellow'),
            formatter_class=ColoredHelpFormatter,
            add_help=False
        )
        command.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show this help message and exit.'
        )
        command.titles('Arguments', 'Options', color='cyan')

        # Add arguments and bind command
        for arg, opt in cmd.arguments.items():
            command.add_argument(arg, **opt)
        command.set_defaults(func=cmd.handle)
        self.commands[name] = command

    def dispatch(self):
        """Use dispatch pattern to invoke class and let it handle the command."""
        try:
            # dispatch
            options = self._parser.parse_args()

            # only dispatch args if available
            if len(vars(options)) > 1:
                options.func(self.app, options)
            else:
                options.func(self.app)
        except AttributeError:
            # no command selected
            print_failure('Please specify a command')
            self._parser.print_help(sys.stderr)
            sys.exit(2)


class ColoredHelpFormatter(argparse.HelpFormatter):

    """HelpFormatter for argparse to create colored output."""

    def _get_help_string(self, action):
        """Color in the default text of the current action."""
        helptext = action.help
        if '%(default)' not in action.help:
            if action.default != '==SUPPRESS==' and action.default:
                # defaulting_nargs = ['?', '*']
                # if action.nargs in defaulting_nargs:
                helptext += colored(' [default: %(default)s]', 'cyan')
        return helptext

    def _format_action_invocation(self, action):
        """Color in action invation (except for option strings)."""
        txt = super(ColoredHelpFormatter, self)._format_action_invocation(
            action
        )
        if action.option_strings:
            return txt

        return colored(txt, 'green')

    def _format_action(self, action):
        """Color in actions."""
        # Color in Options
        if action.option_strings:
            help_position = min(
                self._action_max_length + 2,
                self._max_help_position
            )
            action_width = help_position - self._current_indent - 2
            help_text = ""
            if action.help:
                help_text = self._expand_help(action)

            option_string = ', '.join(action.option_strings)
            option = colored(
                "  {:{width}}".format(option_string, width=action_width),
                'green'
            )
            return '{}  {}\n'.format(option, help_text)

        if isinstance(action, argparse._SubParsersAction):
            # inject new class variable for subcommand formatting
            subactions = action._get_subactions()
            invocations = [
                self._format_action_invocation(a) for a in subactions
            ]
            self._subcommand_max_length = max(len(i) for i in invocations)

        if isinstance(action, argparse._SubParsersAction._ChoicesPseudoAction):
            # format subcommand help line
            subcommand = self._format_action_invocation(action)  # type: str
            width = self._subcommand_max_length
            help_text = ""
            if action.help:
                help_text = self._expand_help(action)

            subcommand = colored(
                "  {:{width}}".format(subcommand, width=width),
                'green'
            )
            return '{}  {}\n'.format(subcommand, help_text)

        elif isinstance(action, argparse._SubParsersAction):
            # process subcommand help section
            msg = ''
            for subaction in action._get_subactions():
                msg += self._format_action(subaction)
            return msg
        else:
            return super(ColoredHelpFormatter, self)._format_action(action)

    def add_usage(self, usage, actions, groups, prefix=None):
        """Color in Usage and change format."""
        if prefix is None:
            prefix = colored('Usage: \n  ', 'cyan')
        return super(ColoredHelpFormatter, self).add_usage(
            usage, actions, groups, prefix)



"""Import classes for easier importing by other packages/modules."""
# flake8: noqa
from .AbstractRepository import AbstractRepository
from .AbstractStore import AbstractStore
from .Migrations import Blueprint, Column, ForeignKey, Migration, Migrator, Schema
from .SQLAlchemy import ColumnFactory, Platform, SQLitePlatform, Store



"""The :py:class:`.Schema` class provides a database agnostic way of manipulating tables.

Tables
======

Creating Tables
---------------
To create a new database table, the :py:meth:`~.Schema.create` method is used.
The :py:meth:`~.Schema.create` method accepts a table name as its argument and returns
a :py:class:`.Blueprint` instance that can be used to define the new table.
When creating the table, you may use any of the :py:class:`.Blueprint` column methods
to define the table's columns::

    with self.schema.create('users') as table:
        table.increments('id')

Checking Existence
------------------
To check if a table or column exist you can use the :py:meth:`~.Schema.has_table` or
:py:meth:`~.Schema.has_column` methods respectively::

    if self.schema.has_table('users'):
        # ...

    if self.schema.has_column('users', 'email'):
        # ...

Renaming / Dropping Tables
--------------------------
To rename an existing database table, use the :py:meth:`~.Schema.rename` method::

    self.schema.rename('from', 'to')

To drop a table, you can use the :py:meth:`~.Schema.drop` or
:py:meth:`~.Schema.drop_if_exists` methods::

    self.schema.drop('users')
    self.schema.drop_if_exists('users')
"""
from experimentum.cli import print_failure
from contextlib import contextmanager


class Schema(object):

    """Database agnostic way of manipulating tables.

    The :py:class:`.Schema` class was inspired by the Laravel Schema Builder
    (https://laravel.com/docs/5.6/migrations#tables).

    Attributes:
        app (App): Main App Class.
        store (AbstractStore): Data Store.
    """

    def __init__(self, app):
        """Set app and store.

        Args:
            app (App): Main App Class.
        """
        self.app = app
        self.store = app.make('store')

    @contextmanager
    def create(self, name):
        """Create a new table blueprint.

        Args:
            name (str): Name of the table.

        Yields:
            Blueprint: New Instance of a table blueprint
        """
        try:
            blueprint = self.app.make('blueprint', name)
            blueprint.create()

            yield blueprint
        except Exception as exc:
            print_failure('Error while creating blueprint: ' + str(exc), 1)

        self._build(blueprint)

    @contextmanager
    def table(self, name):
        """Create a blueprint for an existing table.

        Args:
            name (str): Name of the table

        Yields:
            Blueprint: New Instance of a table blueprint
        """
        try:
            blueprint = self.app.make('blueprint', name)

            yield blueprint
        except Exception as exc:
            print_failure('Error while creating blueprint: ' + str(exc), 1)

        self._build(blueprint)

    def rename(self, old, new):
        """Rename a table.

        Args:
            old (str): Old table name
            new (str): New table name
        """
        self.store.rename(old, new)

    def drop(self, name):
        """Drop a table.

        Args:
            name (str): Name of the table
        """
        self.store.drop(name)

    def drop_if_exists(self, name):
        """Drop a table if it exists.

        Args:
            name (str): Name of the table
        """
        self.store.drop_if_exists(name)

    def has_table(self, table):
        """Check if database has a specific table.

        Args:
            table (str): Table to check existance of
        """
        return self.store.has_table(table)

    def has_column(self, table, column):
        """Check if table has a specific column.

        Args:
            table (str): Table to check
            column (str): Column to check
        """
        return self.store.has_column(table, column)

    def _build(self, blueprint):
        """Build Schema based on the blueprint.

        Args:
            blueprint (Blueprint): Blueprint to build.
        """
        if blueprint.action == 'create':
            self.store.create(blueprint)
        elif blueprint.action == 'alter':
            self.store.alter(blueprint)



# -*- coding: utf-8 -*-
"""Some CLI Helpers."""
from __future__ import print_function
import sys
import six
import shutil
import logging
from termcolor import colored


def get_input(msg, default=None):
    """Get the user input.

    Args:
        msg (str): Message
        default (object, optional): Default to None. Default value if nothing is entered

    Returns:
        object: User Input
    """
    # Default value if skipped
    if default is not None:
        print(colored("› {} [{}]: ".format(msg, default), 'green'), end='')
        return six.moves.input() or default

    # Required input
    print(colored("› {}: ".format(msg), 'green'), end='')
    value = six.moves.input()
    while value == '':
        print(colored("× Please specify a value!", 'red'))
        print(colored("› {}: ".format(msg), 'green'), end='')
        value = six.moves.input()

    return value


def print_progress(iteration, total, prefix='', suffix='', decimals=1, bar_length=100):
    """Call in a loop to create terminal progress bar.

    https://gist.github.com/aubricus/f91fb55dc6ba5557fbab06119420dd6a

    Args:
        iteration (int): current iteration
        total (int): total iterations
        prefix (str, optional): Defaults to ''. Prefix string
        suffix (str, optional): Defaults to ''. Suffix string
        decimals (int, optional): Defaults to 1. Positive number of decimals in percent complete.
        bar_length (int, optional): Defaults to 100. Character length of bar.
    """
    str_format = "{0:." + str(decimals) + "f}"
    percents = str_format.format(100 * (iteration / float(total)))
    filled_length = int(round(bar_length * iteration / float(total)))
    # bar = '█' * filled_length + '-' * (bar_length - filled_length)
    bar = '#' * filled_length + '-' * (bar_length - filled_length)

    sys.stdout.write('\r%s |%s| %s%s %s\n' % (prefix, bar, percents, '%', suffix))

    if iteration == total:
        sys.stdout.write('\n')
    sys.stdout.flush()


def print_failure(msg, exit_code=None):
    """Print a failure message to stder.

    Args:
        msg (str): Failure message.
    """
    try:
        msg_format = '{0}{1}{0}'
        size = shutil.get_terminal_size()[0]  # pylint: disable=no-member
    except Exception:
        msg_format = '{0}{1}{0}\n'
        size = 4

    msg = '{}'.format(msg).center(size, ' ')
    sys.stderr.write(msg_format.format(
        colored(''.center(size, ' '), 'white', 'on_red'),
        colored(msg, 'white', 'on_red', ['bold'])
    ))
    logging.getLogger('experimentum').critical(msg.strip())

    if exit_code is not None:
        sys.exit(exit_code)



"""Column Data Structure to add column modifiers.

Add column modifiers like nullable, default, unsigned to
a column. They support method chaining, so the following is possible::

    table.integer('foo').unsigned().nullable()
"""


class Column(object):

    """Stores attributes the the column."""

    def __init__(self, col_typ, name, parameters):
        """Set Attributes.

        Args:
            col_typ (str): Type of the column.
            name (str): Name of the column.
            parameters (object): Additional parameters for the column.
        """
        self._attributes = {
            'type': col_typ,
            'name': name,
            'parameters': parameters,
            'null': False,
            'default': None,
            'unsigned': False,
        }

    def get(self, attribute, default=None):
        """Get the value of an attribute, with a default if it does not exist.

        Args:
            attribute (str): Attribute you want to get.
            default (object, optional): Defaults to None. Default value if attribute is not found.

        Returns:
            object
        """
        return self._attributes.get(attribute, default)

    def nullable(self):
        """Designate that the column allows NULL values.

        Returns:
            Column: self instance for method chaining.
        """
        self._attributes['null'] = True
        return self

    def default(self, value):
        """Specify a default value for the column.

        Args:
            value (object): The default value

        Returns:
            Column: self instance for method chaining.
        """
        self._attributes['default'] = value
        return self

    def unsigned(self):
        """Set INTEGER column as UNSINGED (MySQL).

        Returns:
            Column: self instance for method chaining.
        """
        self._attributes['unsigned'] = True
        return self

    def __repr__(self):
        """Human-readable representation of the column attributes.

        Returns:
            str
        """
        return '{}'.format(self._attributes)



"""Implementation of the AbstractRepository.

Implements the AbstractRepository interface to use the
SQLAlchemy ORM as a data store.
"""
from sqlalchemy.orm import mapper, relationship
from sqlalchemy.event import listen
from sqlalchemy import or_
from experimentum.Storage import AbstractRepository
import logging


class QueryBuilder(object):

    """Helper Class to build a SQLAlchemy Query.

    Attributes:
        repo (Repository): The Repository to query
        where (list): where condition to build
        __filter_cond (list): Filter Conditions
        __filter_cond_or (list): Filter Conditions that are connected with logical OR
    """

    def __init__(self, repo, where):
        """Set up query builder.

        Args:
            repo (Repository): Repository to query.
            where (list): where condition to build.
        """
        self.__filter_cond = []
        self.__filter_cond_or = []

        self.repo = repo

        if not isinstance(where, list):
            where = []
        if len(where) == 0 or not isinstance(where[0], list):
            where = [where]

        self.where = where

    def build(self, query):
        """Build the query.

        Args:
            query (sqlalchemy.orm.query.Query): Current query

        Returns:
            sqlalchemy.orm.query.Query: Final query
        """
        query = self.build_where(query)
        return query

    def build_where(self, query):
        """Build the query with the where conditions.

        Args:
            query (sqlalchemy.orm.query.Query): Current Query

        Returns:
            sqlalchemy.orm.query.Query: Query with where conditions applied.
        """
        for cond in self.where:
            # ['id', 2] => WHERE id == 2
            if len(cond) == 2:
                self.__filter_cond.append(getattr(self.repo, cond[0]) == cond[1])
            # ['or', 'id', 2] => WHERE id == 2 OR ...
            elif len(cond) == 3 and cond[0] == 'or':
                self.__filter_cond_or.append(getattr(self.repo, cond[1]) == cond[2])
            # ['id', '!=', 2] => WHERE id != 2
            # ['or', 'id', '==', 2] => WHERE id == 2 OR
            elif (len(cond) == 3 and cond[0] != 'or') or (len(cond) == 4 and cond[0] == 'or'):
                operator = len(cond) - 2
                left = operator - 1
                right = operator + 1

                if cond[operator] == '!=':
                    self.__filter_cond.append(getattr(self.repo, cond[left]) != cond[right])
                elif cond[operator] == '==':
                    self.__filter_cond.append(getattr(self.repo, cond[left]) == cond[right])
                elif cond[operator] == '>':
                    self.__filter_cond.append(getattr(self.repo, cond[left]) > cond[right])
                elif cond[operator] == '<':
                    self.__filter_cond.append(getattr(self.repo, cond[left]) < cond[right])
                elif cond[operator] == '>=':
                    self.__filter_cond.append(getattr(self.repo, cond[left]) >= cond[right])
                elif cond[operator] == '=<':
                    self.__filter_cond.append(getattr(self.repo, cond[left]) <= cond[right])

        return query.filter(*self.__filter_cond).filter(or_(*self.__filter_cond_or))


class Repository(AbstractRepository):

    """Implementation of the AbstractRepository Interface.

    This implementation uses the SQLAlchemy ORM to implement the
    *Repository Design Pattern*. It uses the SQLAlchemy ``mapper`` function
    to map the Table Schema and the a Repository class. It also uses the
    ``listen`` method to hook up the before_* and after_* events.
    """
    _events = {
        'before_insert': lambda m, conn, target: target.before_insert(),
        'after_insert': lambda m, conn, target: target.after_insert(),
        'before_update': lambda m, conn, target: target.before_update(),
        'after_update': lambda m, conn, target: target.after_update()
    }

    def create(self):
        """Save the repository content in your data store.

        Returns:
            Repository: Self Instance
        """
        self.store.session.add(self)
        self.store.session.commit()

        return self

    def update(self):
        """Update the repository content in your data store.

        Returns:
            Repository: Self Instance.
        """
        self.store.session.commit()
        return self

    def delete(self):
        """Delete the repository content from your data store.

        Returns:
            Repository: Self Instance.
        """
        self.store.session.query(self.__class__).filter(self.__class__.id == self.id).delete()
        self.store.session.commit()
        return self

    @classmethod
    def get(cls, where=None):
        """Get all entries which satisfy a specific condition from your data store.

        Args:
            where (list, optional): Defaults to None. Where Condition

        Returns:
            list: List of items which satisfy the condition.
        """
        query = cls.store.session.query(cls)
        builder = QueryBuilder(cls, where)

        return builder.build(query)

    @classmethod
    def first(cls, where=None):
        """Get first entry which satisfy a specific condition from your data store.

        Args:
            where (list, optional): Defaults to None. Where Condition

        Returns:
            AbstractRepository: Item which satisfies the condition.
        """
        return cls.get(where).first()

    @classmethod
    def find(cls, id):
        """Find an entry of this repository based on its id.

        Args:
            id (int): ID to search for.

        Returns:
            AbstractRepository: Item which the concrete id
        """
        return cls.first(where=['id', id])

    @classmethod
    def all(cls):
        """Get all entries for this specific repository from your data store.

        Returns:
            list: List of all entires
        """
        return cls.store.session.query(cls).all()

    @staticmethod
    def mapping(cls, store):
        """Map data store content to repository classes.

        Example:
        Map a *User* Table in the data store to the *UserRepository*,
        which reflects the attributes of the User Table schema::

            Repository.mapping(UserRepository, store)

        Args:
            cls (AbstractRepository): Repository to map
            store (AbstractStore): Storage that is use
        """
        # set store
        cls.store = store

        # Map classes with relationships
        try:
            relationships = {}
            for key, relation in cls.__relationships__.items():
                if len(relation) == 2:
                    relationships[key] = relationship(relation[0], **relation[1])
                else:
                    relationships[key] = relationship(relation[0])

                # recursively map relations first otherwise it won't work!
                Repository.mapping(relation[0], store)

            Repository.map_to_table(
                cls, cls, cls.store.meta.tables.get(cls.__table__), properties=relationships
            )
        except Exception:
            logging.getLogger('experimentum').warning('Could not map table: ' + cls.__table__)

        # Listen to events
        for event, callback in cls._events.items():
            listen(cls, event, callback)

    @staticmethod
    def map_to_table(cls, repo, table, properties):
        """Get the SQLAlchemy mapper, i.e. map the tables to classes.

        Args:
            repo (Repository): Repository to map
            table (sqlalchemy.schema.table): Table to map
            properties (object): relationships to map
        """
        mapper(repo, table, properties=properties)



# -*- coding: utf-8 -*-
"""The Performance module lets you easily measure the performance.

It measures *execution time and memory consumption* of your python script
and lets you add messages to each measuring point for a more detailed overview.

Example:

.. code-block:: python

    performance = Performance()
    with performance.point('Task A') as point:
        # Do some heavy work
        point.message('Database query insert xy')

        with performance.point('Subtask A1') as subpoint:
            # Do some heavy work

        with performance.point('Subtask A2') as subpoint:
            subpoint.message('Database query insert xy')

    performance.results()  # print results table
"""
from __future__ import print_function
from contextlib import contextmanager
from timeit import time
from termcolor import colored
import collections
import math
import psutil
import os
import tabulate
tabulate.PRESERVE_WHITESPACE = True


def memory_usage():
    """Return the memory usage of the current process.

    Note:
        Thanks to Fabian Pedregosa for his overview of different ways to
        determine the memory usage in python.
        http://fa.bianp.net/blog/2013/different-ways-to-get-memory-consumption-or-lessons-learned-from-memory_profiler

    Returns:
        float: used bytes.
    """
    process = psutil.Process(os.getpid())

    # return process.get_memory_info()[0] / float(2 ** 20)
    # return process.memory_info()[0]
    return process.memory_full_info().uss


def _to_df(point, level=0):
    """Transfrom point to dataframe dict layout.

    Args:
        point (dict): Point
        level (int, optional): Defaults to 0. Level

    Returns:
        dict
    """
    data = {
        'Label': [point['label']],
        'Time': [point['difference_time']],
        'Memory': [point['difference_memory']],
        'Peak Memory': [point['peak_memory']],
        'Level': [level],
        'Type': ['point'],
        'ID': [point['id']],
        'Key': ['{}_{}_{}'.format(point['id'], level, point['label'])]
    }

    for msg in point['messages']:
        data['Label'].append(msg[1])
        data['Time'].append(msg[0])
        data['Memory'].append(point['difference_memory'])
        data['Peak Memory'].append(point['peak_memory'])
        data['Level'].append(level)
        data['Type'].append('message')
        data['ID'].append(point['id'])
        data['Key'].append('{}_{}_{}'.format(point['id'], level, msg[1]))

    for subpoint in point['subpoints']:
        _point = subpoint.to_dict()
        result = _to_df(_point, level+1)
        data['Label'].extend(result['Label'])
        data['Time'].extend(result['Time'])
        data['Memory'].extend(result['Memory'])
        data['Peak Memory'].extend(result['Peak Memory'])
        data['Level'].extend(result['Level'])
        data['Type'].extend(result['Type'])
        data['ID'].extend(result['ID'])
        data['Key'].extend(result['Key'])

    return data


def _calc_metrics(row):
    """Calculate metrics for time and memory.

    Args:
        row (list): List of points

    Returns:
        set: Mean Time, STD Time, Mean Memory, STD Memory
    """
    times = list(zip(*[points['Time'] for points in row]))
    mean_time = list(map(Performance.mean, times))
    std_time = list(map(Performance.standard_deviation, times))

    memory = list(zip(*[points['Memory'] for points in row]))
    mean_memory = list(map(Performance.mean, memory))
    std_memory = list(map(Performance.standard_deviation, memory))

    return list(zip(mean_time, std_time, mean_memory, std_memory))


def _transform_points(points, attrs):
    """Transform points to another format for better usability.

    Args:
        points (dict): Points
        attrs (list): list of point attributes

    Returns:
        list: List of Points in the format of [{'label': 'foo', ...}, {'label': 'bar'}]
    """
    items = zip(*[points[k] for k in attrs])
    return [
        {
            key.lower().replace(' ', '_'): item[idx] for idx, key in enumerate(attrs)
        }
        for item in items
    ]


class Formatter(object):

    """Format performance values to a human-readable format."""

    def memory_to_human(self, _bytes, unit='auto', decimals=2):
        """Transform used memory into a better suited unit.

        Args:
            _bytes (float): Used Bytes.
            unit (str, optional): Defaults to 'auto'. Unit to transform seconds into.
            decimals (int, optional): Defaults to 2. To which decimal point is rounded to.

        Returns:
            string: used memory suffixed with unit identifer
        """
        if _bytes <= 0:
            return '0.00 KB'

        # If decimals is less than 0
        decimals = 2 if decimals < 0 else decimals

        value = 0
        units = {
            'B': 0, 'KB': 1, 'MB': 2, 'GB': 3, 'TB': 4, 'PB': 5, 'EB': 6, 'ZB': 7, 'YB': 8
        }

        if _bytes > 0:
            # Generate automatic prefix by bytes if wrong prefix given
            if unit not in units:
                power = math.floor(math.log(_bytes) / math.log(1024))
                unit = [k for k, v in units.items() if v == power][0]

            # Calculate byte value by prefix
            value = (_bytes / pow(1024, math.floor(units[unit])))

        return self.format_number(value, decimals, unit)

    def time_to_human(self, seconds, unit='auto', decimals=2):
        """Transform time into a better suited unit.

        Args:
            seconds (float): Time seconds.
            unit (str, optional): Defaults to 'auto'. Unit to transform seconds into.
            decimals (int, optional): Defaults to 2. To which decimal point is rounded to.

        Raises:
            TypeError: if an unknown unit is passed

        Returns:
            string: time suffixed with unit identifer
        """
        if unit == 'auto':
            if seconds >= 1:
                unit = 's'
            elif seconds >= 0.001:
                unit = 'ms'
            else:
                unit = u'\u03BCs'

        if unit == u'\u03BCs':
            return self.format_number(seconds * 1000000, decimals, unit)
        elif unit == 'ms':
            return self.format_number(seconds * 1000, decimals, unit)
        elif unit == 's':
            return self.format_number(seconds * 1, decimals, unit)
        else:
            raise TypeError('Performance format {} does not exist.'.format(unit))

    def print_table(self, points, tablefmt='psql'):
        """Print performance table in human-readable format.

        Args:
            points (list): Measuring Points.
            tablefmt (str, optional): Defaults to 'psql'. Table format for :mod:`tabulate`
        """
        print(self.get_table(points, tablefmt))

    def get_table(self, points, tablefmt='psql'):
        """Print performance table in human-readable format.

        Args:
            points (list): Measuring Points.
            tablefmt (str, optional): Defaults to 'psql'. Table format for :mod:`tabulate`

        Returns:
            str: Performance table
        """
        data = []
        headers = [
           colored('Label', 'cyan', attrs=['bold']),
           colored('Time', 'cyan', attrs=['bold']),
           colored('Memory', 'cyan', attrs=['bold']),
           colored('Peak Memory', 'cyan', attrs=['bold'])
        ]

        # Color items and build data list
        for row in points:
            label = {'format': u'› {}', 'attrs': ['bold']}
            time = {
                'val': self.time_to_human(row['mean_time']),
                'std': self.time_to_human(row['std_time']),
                'attrs': ['bold']
            }
            memory = {
                'val': self.memory_to_human(row['mean_memory']),
                'std': self.memory_to_human(row['std_memory']),
                'peak': self.memory_to_human(row['peak_memory']),
                'frmt': u'{} (± {})',
                'attrs': ['bold']
            }

            if row['type'] == 'message':
                label = {'format': '  {}', 'attrs': ['dark']}
                memory['frmt'] = '--'
                memory['peak'] = '--'

            data.append([
                colored(label['format'].format(row['label']), attrs=label['attrs']),
                u'{} (± {})'.format(colored(time['val'], attrs=time['attrs']), time['std']),
                memory['frmt'].format(colored(memory['val'], attrs=memory['attrs']), memory['std']),
                colored(memory['peak'], attrs=memory['attrs'])
            ])

        return tabulate.tabulate(data, headers, tablefmt=tablefmt)

    @staticmethod
    def format_number(value, decimals, unit):
        """Round a number and add a unit.

        Args:
            value (float): Number to round.
            decimals (int): Numer of decimals places.
            unit (str): Unit to append

        Returns:
            string
        """
        _format = u'{:.%sf} {}' % decimals
        return _format.format(value, unit)


class Point(object):

    """Measuring Point Data Structure.

    Keeps track of execution time and memory consumption. You can also append
    messages to a point to keep track of different events. Each message contains
    the time since the last message was logged and the message content itself.

    Attributes:
        label (str): Label of the point.
        id (int): Id to keep track of same points when iterating.
        start_time (float): Startpoint Timestamp.
        stop_time (float): Endpoint Timestamp.
        start_memory (int): Memory consumption on start.
        stop_memory (int): Memory consumption on end.
        messages (list): List of optional messages.
        subpoints (list): List of optional subpoints.
    """

    def __init__(self, label, iter_id=None):
        """Set the current time and memory consumption and default values for other attributes.

        Args:
            label (str): Label of the point.
            iter_id (int): Id to keep track of same points when iterating.
        """
        self.label = label
        self.id = iter_id
        self.start_time = time.time()
        self.stop_time = 0
        self.start_memory = memory_usage()
        self.stop_memory = 0
        self.messages = []
        self.subpoints = []

        self._last_msg = self.start_time

    def message(self, msg):
        """Set a message associated with the point.

        Args:
            msg (str): Enter message
        """
        msg_time = time.time()
        self.messages.append([msg_time - self._last_msg, msg])
        self._last_msg = msg_time

    def to_dict(self):
        """Return all measured attributes and messages as a dictionary.

        Returns:
            dict: measured attributes and messages.
        """
        return {
            'label': self.label,
            'id': self.id,
            'start_time': self.start_time,
            'stop_time': self.stop_time,
            'difference_time': self.stop_time - self.start_time,
            'start_memory': self.start_memory,
            'stop_memory': self.stop_memory,
            'difference_memory': self.stop_memory - float(self.start_memory),
            'peak_memory': max(self.stop_memory, self.start_memory),
            'messages': self.messages,
            'subpoints': self.subpoints
        }

    def to_df(self):
        """Transform point to dataframe.

        Returns:
            dict: Dataframe
        """
        return _to_df(self.to_dict())


class Performance(object):

    """Easily measure the performance of your python scripts.

    Attributes:
        points (list): List of measuring points
        iteration (int): Number of current iteration
        formatter (Formatter): Formatter to output human readable results
    """

    def __init__(self):
        """Set measuring points list and default formatter."""
        self.points = []
        self.iteration = 0
        self.set_formatter(Formatter())

    def set_formatter(self, formatter):
        """Set a formatter for human readable output.

        Args:
            formatter (Formatter): Human-readable output formatter
        """
        self.formatter = formatter

    def iterate(self, start, stop):
        """Iterate over multiple performance points to later calculate avg and standard deviation.

        Args:
            start (int): Start at
            stop (int): Stop at

        Yields:
            int: current iteration
        """
        current = start
        while current <= stop:
            self.iteration = 0
            yield current
            current += 1

    @contextmanager
    def point(self, label='Point'):
        """Set measuring point with or without a label.

        Keyword Arguments:
            label (str, optional): Defaults to 'Point'. Enter point label

        Example::

            with performance.point('Task A') as point:
                # do something
                point.message('Some Message)

        Yields:
            Point: new measuring point
        """
        try:
            self.iteration += 1
            point = Point(label, self.iteration)

            if len(self.points) and self.points[-1].stop_time == 0:
                self.points[-1].subpoints.append(point)
            else:
                self.points.append(point)

            yield point

        except Exception as exc:
            print('Exception: {}'.format(exc))
        finally:
            point.stop_time = time.time()
            point.stop_memory = memory_usage()

    def export(self, metrics=False):
        """Export the measuring points as a dictionary.

        Args:
            metrics (bool, optional). Defaults to False. Whether or not metrics should be calculated

        Returns:
            dict: Measuring points
        """
        # Build grouped Dataframe of points
        frames = [point.to_df() for point in self.points]
        result = collections.OrderedDict()
        for frame in frames:
            key = '|'.join(frame['Key'])
            if key in result:
                result[key].append(frame)
            else:
                result[key] = [frame]

        # Transform to ouput format
        data = []
        for row in result.values():
            # Only last iteration
            points = _transform_points(
                row[-1], ['Label', 'Level', 'Type', 'Time', 'Memory', 'Peak Memory']
            )

            # Calculate metrics for time and memory
            if metrics:
                metrics = _calc_metrics(row)

                # add metrics to each point
                for idx, point in enumerate(points):
                    point['mean_time'] = metrics[idx][0]
                    point['std_time'] = metrics[idx][1]
                    point['mean_memory'] = metrics[idx][2]
                    point['std_memory'] = metrics[idx][3]

            # add points to data list
            data.extend(points)

        return data

    def results(self):
        """Print the performance results in a human-readable format."""
        self.formatter.print_table(self.export(metrics=True))

    # Mean and Standard Deviation
    @staticmethod
    def mean(values):
        """Calculate Mean of values.

        Args:
            values (list): List of values

        Returns:
            float: Mean value
        """
        return sum(values) / float(len(values))

    @staticmethod
    def standard_deviation(numbers):
        """Calculate the standard deviation.

        Args:
            numbers (list): List of numbers

        Returns:
            float: Standard Deviation
        """
        mean = Performance.mean(numbers)
        return math.sqrt(
            Performance.mean([math.pow(number - mean, 2) for number in numbers])
        )



"""Provides the main entry point and some utilities for running experiments.

Import :py:mod:`.App`, :py:mod:`Experiment` and :py:mod:`.Performance` so that
they can be easly imported by other packages/modules.
"""
# flake8: noqa
from .Performance import Performance
from .Experiment import Experiment, Script
from .DataBag import DataBag
from .App import App



"""The Config Loader module loads json config files.

The :py:mod:`.Loader` loads all json files from a folder
and stores the content in the :py:class:`.Config` class.
"""

import logging
import json
import glob
import os
import io


class Loader(object):

    """Load the configuration files.

    Attributes:
        config_path (str): Path to config files.
        config (Config): Config holder class
    """

    def __init__(self, config_path, config):
        """Load the configuration items.

        Arguments:
            config_path (str): Path to config files.
            config (Config): Config holder class.
        """
        self.config_path = config_path
        self.config = config

    def load_config_files(self):
        """Load the configuration items from all of the files.

        Raises:
            Exception: if there is no app.json config.
        """
        files = self.get_config_files()

        if 'app' not in files:
            logging.critical('Unable to load the "app" configuration file.')
            raise Exception('Unable to load the "app" configuration file.')

        for key, path in files.items():
            with io.open(path, encoding='utf-8') as filehandler:
                self.config.set(key, json.load(filehandler))

    def get_config_files(self):
        """Get all of the configuration files for the application.

        Returns:
            dict: The config items.
        """
        files = glob.glob(os.path.join(self.config_path, '*.json'))

        return {
           self.config_key(file): os.path.realpath(file) for file in files
        }

    @classmethod
    def config_key(self, fname):
        """Get the config key based on the filename.

        Arguments:
            fname (str): Name of the config file.

        Returns:
            str: key
        """
        return os.path.splitext(os.path.basename(fname))[0]



"""Concrete Implementation of a data store based on SQLAlchemy.

Uses the SQLAlchemy ORM to implement the :py:mod:`.AbstractStore` interface.
"""
from experimentum.Storage import AbstractStore
from experimentum.Storage.SQLAlchemy import SQLitePlatform, Platform, ColumnFactory
from sqlalchemy import inspect, MetaData, Table
from sqlalchemy.orm import sessionmaker


class Store(AbstractStore):

    """Concrete Implementation of a data store based on SQLAlchemy.

    Attributes:
        app (App): Framework App class.
        engine (sqlalchemy.engine.Engine): Defaults to None. Database engine.
        meta (sqlqlchemy.schema.MetaData): Defaults to None. Schema Meta Data.
        platform (Platform): Basic SQL Statements because SQLAlchemy could not handle everything.
        sqlite_platform (SQLitePlatform): SQLite specific sql statements.
    """

    def __init__(self, app):
        """Set app and load platform handlers.

        Args:
            app (App): Framework App class.
        """
        self.app = app
        self.platform = Platform()
        self.sqlite_platform = SQLitePlatform()
        self.factory = ColumnFactory()
        self.engine = None
        self.meta = None
        self.session = None

    def set_engine(self, engine):
        """Set database engine, metadata store, and platform specific handlers.

        Args:
            engine (sqlalchemy.engine.Engine): Database engine
        """
        self.engine = engine
        self.meta = MetaData()
        self.meta.bind = self.engine
        self.meta.reflect()  # Reflecting All Tables at Once
        self.platform.set_engine(self.engine, self.meta)
        self.sqlite_platform.set_engine(self.engine, self.meta)

        # Create session
        session = sessionmaker(bind=engine)
        self.session = session()

    def has_table(self, table):
        """Check if the data store has a specific table.

        Args:
            table (str): Name of the Table

        Returns:
            boolean
        """
        inspector = inspect(self.engine)
        return table in inspector.get_table_names()

    def has_column(self, table, column):
        """Check if a table has a specific column.

        Args:
            table (str): Name of the table
            column (str): Name of the column

        Returns:
            boolean
        """
        inspector = inspect(self.engine)
        columns = inspector.get_columns(table)
        for col in columns:
            if col.get('name') == column:
                return True

        return False

    def drop(self, name, checkfirst=False):
        """Drop a table from the data store.

        Notes:
            http://docs.sqlalchemy.org/en/latest/core/metadata.html#creating-and-dropping-database-tables

        Args:
            name (str): Name of the table
            checkfirst (bool, optional): Defaults to False. Whether to check if exists first.
        """
        table = Table(name, self.meta)
        table.drop(self.engine, checkfirst=checkfirst)

    def drop_if_exists(self, name):
        """Drop a table from the datastore if it exists.

        Notes:
            http://docs.sqlalchemy.org/en/latest/core/metadata.html#creating-and-dropping-database-tables

        Args:
            name (str): Name of the table
        """
        table = Table(name, self.meta)
        table.drop(self.engine, checkfirst=True)

    def rename(self, old, new):
        """Rename a table.

        Args:
            old (str): Old table name
            new (str): New table name
        """
        self.engine.execute(self.platform.get_rename_sql(old, new))

    def create(self, blueprint):
        """Create a new Table with Columns and indexes.

        Notes:
            http://docs.sqlalchemy.org/en/latest/core/metadata.html#creating-and-dropping-database-tables

        Args:
            blueprint (Blueprint): The Blueprint to create the table
        """
        # Collection of columns and indexes
        data = self.factory.get_columns_and_indexes(blueprint)
        data['columns'].extend(data['indexes'])

        # Create table
        table = Table(blueprint.table, self.meta, *data['columns'], extend_existing=True)
        table.create(self.engine, checkfirst=True)

    def alter(self, blueprint):
        """Alter Schema of the table.

        Args:
            blueprint (Blueprint): The new blueprint with updated schema.
        """
        # Get columns and indexes
        data = self.factory.get_columns_and_indexes(blueprint)

        # Alter SQLite Database
        if self.platform.is_sqlite():
            data['columns'].extend(data['indexes'])
            self.sqlite_platform.alter_table(blueprint.table, data['columns'], blueprint.dropped)
            return

        # Alter SQL Database -> Add Columns and keys
        table = Table(blueprint.table, self.meta, extend_existing=True)
        for column in data['columns']:
            if not self.has_column(blueprint.table, column.name):
                table.append_column(column)
                self.engine.execute(self.platform.get_add_column_sql(blueprint.table, column))

            # Foreign Kyes and Primary Keys
            key_sql = self.platform.get_key_sql(blueprint.table, column)
            for sql in key_sql:
                self.engine.execute(sql)

        # Create indexes
        for index in data['indexes']:
            table.append_constraint(index)
            index.create()

        # Drop indexes
        for drop_idx in blueprint.dropped['indexes']:
            self.engine.execute(self.platform.get_drop_key_constraint_sql(
                blueprint.table,
                table.columns[drop_idx['col']],
                drop_idx['type'],
                drop_idx['name']
            ))

        # Drop Columns
        drop_sql = self.platform.get_drop_columns_sql(
            blueprint.table,
            list(set(table.columns.keys()) & set(blueprint.dropped['columns']))
        )
        for sql in drop_sql:
            self.engine.execute(sql)



"""In-Memory Container to easily store experiment results on the fly.

Adding Entries
--------------
Add the entry `{'foo': {'bar': {'baz': 42} }}` to the DatBag via a dot-notation supported key::

    DataBag.add('foo.bar.baz', 42)

Mergins Entries
---------------
Merging dictionary and list entires::

    DataBag.add('foo.bar', [1, 2, 3])
    DataBag.add('foo.baz', {'a': 1, 'b': 2})
    DataBag.merge('foo.bar', [2, 4, 8])
    DataBag.merge('foo.baz', {'a': 42, 'c': 3})

    # Results in
    # {
    #   'foo': {
    #       'bar': [1, 2, 3, 2, 4, 8],
    #       'baz': {'a': 42, 'b': 2, 'c': 3}
    #   }
    # }


Getting Entries
---------------
Add the entry `baz` of the dictionary `{'foo': {'bar': {'baz': 42} }}` in the DatBag via a
dot-notation supported key::

    data = DataBag.get('foo.bar.baz')  # 42

The DataBag also provides a default value if the key does not exist
    data = DataBag.get('a.b.c', 'default value')  # 'default value'


Deleting Entries
----------------
Deleting entries works just like adding/getting entries::

    DataBag.delete('foo.bar.baz')  # {'foo': {'bar': {}}}


Flushing Entries
----------------
Flushing the DataBag means that you get all the entires (or just a specific key)
and then clear the content::

    DataBag.add('foo.bar.baz', 42)
    print(DataBag.flush('foo.bar.baz))  # prints: 42, contains: {'foo': {'bar': {}}}
    print(DataBag.flush())              # prints: {'foo': {'bar': {}}}, contains: {}}

"""


def set_to(items, data, value):
    """Recursively set value to dictionary deep key.

    Args:
        items (list): List of dictionary keys
        data (dict): Portion of dictionary to operate on
        value (object): Value to set for key.
    """
    item = items.pop(0)

    # Traverse dictionary to last item key
    if items:
        if item not in data:
            data[item] = {}
        set_to(items, data[item], value)
    # Set Value on last key
    else:
        # If the key already exists and is not a list -> transfrom to list and append value
        if item in data:
            if not isinstance(data[item], list):
                data[item] = [data[item]]
            data[item].append(value)
        else:
            data[item] = value


def get_from(items, data):
    """Recursively get value from dictionary deep key.

    Args:
        items (list): List of dictionary keys
        data (dict): Portion of dictionary to operate on

    Returns
        object: Value from dictionary

    Raises:
        KeyError: If key does not exist
    """
    if not isinstance(data, dict):
        raise KeyError

    item = items.pop(0)
    data = data[item]

    return get_from(items, data) if items else data


def del_key(items, data):
    """Recursively remove deep key from dict.

    Args:
        items (list): List of dictionary keys
        data (dict): Portion of dictionary to operate on

    Raises:
        KeyError: If key does not exist
    """
    if not isinstance(data, dict):
        raise KeyError

    item = items.pop(0)
    if items and item in data:
        return del_key(items, data[item])
    elif not items and item in data:
        del data[item]
    elif not items and item not in data:
        raise KeyError(item)


class DataBag(object):

    """In-Memory Container to easily store experiment results on the fly."""
    _data = {}

    @classmethod
    def add(cls, key, value):
        """Add an item to the DatBag via a dot-notation supported key.

        Args:
            key (str): Key to set (i.e. 'foo' or 'foo.bar.baz' for a deep key)
            value (object): Value to set for key.
        """
        set_to(key.split('.'), cls._data, value)

    @classmethod
    def get(cls, key, default=None):
        """Get an item from the DataBag via a dot-notation supported key.

        Args:
            key(str): Key to get (i.e. 'foo' or 'foo.bar.baz' for a deep key)
            default(object, optional): Defaults to None. Default if key does not exist

        Returns:
            object: Value of the key
        """
        try:
            return get_from(key.split('.'), cls._data)
        except KeyError:
            return default

    @classmethod
    def delete(cls, key):
        """Delete an item from the DataBag via a dot-notation supported key.

        Args:
            key(str): Key to delete (i.e. 'foo' or 'foo.bar.baz' for a deep key)

        Returns:
            int: Statuscode, 1: okay, -1: Key not found
        """
        try:
            del_key(key.split('.'), cls._data)
        except KeyError:
            return -1

        return 1

    @classmethod
    def merge(cls, key, data):
        """Merge two dicts or lists.

        For example:
            - {'a': 1, 'b': 2} + {'a': 42, 'c': 3}  => {'a': 42, 'b': 2, 'c': 3}
            - ['a', 'b', 'c']  + ['a', 'd', 'e']    => ['a', 'b', 'c', 'a', 'd', 'e']

        Args:
            key (str): Key to set (i.e. 'foo' or 'foo.bar.baz' for a deep key)
            data (dict|list): Value to update/extend for key.
        """
        items = cls.get(key)

        if items:
            if isinstance(items, dict):
                items.update(data)
            elif isinstance(items, list):
                items.extend(data)

    @classmethod
    def all(cls):
        """Return content of the DataBag.

        Returns:
            dict: DataBag content.
        """
        return cls._data

    @classmethod
    def clear(cls):
        """Clear DataBag content."""
        cls._data = {}

    @classmethod
    def flush(cls, key=None, default=None):
        """Flush a specific key or all of the databag (i.e. return and delete).

        Args:
            key(str): Key to set(i.e. 'foo' or 'foo.bar.baz' for a deep key)
            default(object, optional): Defaults to None. Default if key does not exist

        Returns:
            object: Value of the key
        """
        # flush specific key
        if key:
            data = cls.get(key, default)
            cls.delete(key)
            return data

        # flush all
        data = cls.all()
        cls.clear()
        return data



"""SQL Platform specific code.

Generate SQL queries for altering tables and columns because
SQLAlchemy mostly does not support these types of operations.
"""
from sqlalchemy.dialects.mysql.base import MySQLDialect
from sqlalchemy.dialects.mssql.base import MSDialect
from sqlalchemy.dialects.sqlite.base import SQLiteDialect
from sqlalchemy.dialects.postgresql.base import PGDialect


class Platform(object):

    """Gettings some sql platform specific queries.

    Attributes:
        engine (sqlalchemy.engine.Engine): Database Engine
        meta (sqlalchemy.schema.MetaData): Schema Metadata
    """

    @staticmethod
    def get_rename_sql(old, new):
        """Get the sql for renaming a table.

        Args:
            old (str): old table name
            new (str): new table name

        Returns:
            str
        """
        return 'ALTER TABLE {} RENAME TO {};'.format(old, new)

    @staticmethod
    def get_drop_columns_sql(table, columns):
        """Get a list of sql commands for dropping columns.

        Args:
            table (str): name of the table
            columns (list): list of column names

        Returns:
            list
        """
        return ['ALTER TABLE {} DROP COLUMN {};'.format(table, column) for column in columns]

    @staticmethod
    def get_foreign_key_action_sql(foreign_key):
        """Get ONUPDATE/ONDELETE actions.

        Args:
            foreign_key (sqlalchemy.schema.ForeignKey): ForeignKey to get action of.

        Returns:
            dict
        """
        actions = {
            'on_delete': '',
            'on_update': ''
        }

        if foreign_key.ondelete:
            actions['on_delete'] = 'ON DELETE {}'.format(foreign_key.ondelete)
        if foreign_key.onupdate:
            actions['on_update'] = 'ON UPDATE {}'.format(foreign_key.onupdate)

        return actions

    def __init__(self):
        """Initialize platform."""
        self.engine = None
        self.meta = None

    def set_engine(self, engine, meta):
        """Set engine and meta.

        Args:
            engine (sqlalchemy.engine.Engine): Database Engine
            meta (sqlalchemy.schema.MetaData): Schema Metadata
        """
        self.engine = engine
        self.meta = meta

    def is_sqlite(self):
        """Check if current database driver is SQLite.

        Returns:
            boolean
        """
        return isinstance(self.engine.dialect, SQLiteDialect)

    def is_mysql(self):
        """Check if current database driver is MySQL.

        Returns:
            boolean
        """
        return isinstance(self.engine.dialect, MySQLDialect)

    def is_mssql(self):
        """Check if current database driver is MSSQL.

        Returns:
            boolean
        """
        return isinstance(self.engine.dialect, MSDialect)

    def is_postgresql(self):
        """Check if current database driver is PostgreSQL.

        Returns:
            boolean
        """
        return isinstance(self.engine.dialect, PGDialect)

    def get_add_column_sql(self, table, column):
        """Get SQL for adding a column to a table.

        Notes:
            - https://stackoverflow.com/questions/7300948/add-column-to-sqlalchemy-table#17243132
            - http://docs.sqlalchemy.org/en/latest/core/compiler.html

        Args:
            table (str): Name of the table
            column (sqlalchemy.schema.Column): Column to add

        Returns:
            str
        """
        column_name = column.compile(dialect=self.engine.dialect)
        column_type = column.type.compile(self.engine.dialect)

        # postgresql does not like the table.column syntax
        if self.is_postgresql():
            column_name = column.name

        return 'ALTER TABLE {table} ADD COLUMN {column} {type};'.format(
            table=table,
            column=column_name,
            type=column_type
        )

    def get_key_sql(self, table, column):
        """Get SQL for adding foreign key and primary key constraints.

        Args:
            table (str): Name of the table
            column (sqlalchemy.schema.Column): Column to add contraints on

        Returns:
            list: List of strings with SQL commands
        """
        sql = []

        # add foreign keys
        for fkey in column.foreign_keys:
            actions = self.get_foreign_key_action_sql(fkey)
            sql.append(
                'ALTER TABLE {} ADD CONSTRAINT {} FOREIGN KEY ({}) REFERENCES {}({}) {} {};'.format(
                    table,
                    fkey.name.replace('.', '__'),
                    column.name,
                    fkey.target_fullname.split('.')[0],
                    fkey.target_fullname.split('.')[1],
                    actions['on_delete'],
                    actions['on_update']
                )
            )

        # add primary key if needed
        if column.primary_key:
            sql.append('ALTER TABLE {} ADD PRIMARY KEY({});'.format(
                table,
                column.compile(dialect=self.engine.dialect)
            ))

        return sql

    def get_drop_key_constraint_sql(self, table, column, idx_type, idx_name):
        """Get a drop key constraint sql query.

        Args:
            table (str): Name of table
            column (sqlalchemy.schema.Column): Column with the key
            idx_type (str): Type of index, i.e. primary, foreign, unique or index
            idx_name (str): Name of the index

        Returns:
            str
        """
        column_name = column.compile(dialect=self.engine.dialect)
        column_type = column.type.compile(self.engine.dialect)

        if idx_type == 'primary':
            if self.is_mysql():
                return 'ALTER TABLE {0} CHANGE {1} {1} {2} UNSIGNED NOT NULL; ' \
                    'ALTER TABLE {0} DROP PRIMARY KEY;'.format(table, column_name, column_type)

            return 'ALTER TABLE {} DROP CONSTRAINT {};'.format(table, idx_name)
        elif idx_type == 'unique':
            return 'ALTER TABLE {} DROP {} {};'.format(
                table,
                'INDEX' if self.is_mysql() else 'CONSTRAINT',
                idx_name
            )
        elif idx_type == 'foreign':
            return 'ALTER TABLE {} DROP {} {};'.format(
                table,
                'FOREIGN KEY' if self.is_mysql() else 'CONSTRAINT',
                idx_name
            )
        elif idx_type == 'index':
            if self.is_mssql():
                return 'DROP INDEX {} ON {};'.format(idx_name, table)
            elif self.is_mysql():
                return 'ALTER TABLE {} DROP INDEX {};'.format(table, idx_name)

            return 'DROP INDEX {};'.format(idx_name)



"""Implementation of the Storage based on SQLAlchemy."""
# flake8: noqa
from .Repository import Repository
from .Platform import Platform
from .SQLitePlatform import SQLitePlatform
from .ColumnFactory import ColumnFactory
from .Store import Store



r"""ForeignKey data structure to configure foreign key attributes.

Supports method chaining to easily configure foreign keys, like::

    table.foreign('user_id').references('id').on('users')\
        .on_delete('cascade')\
        .on_update('cascade')
"""


class ForeignKey(object):

    """Stores attributes for a Foreign Key."""

    def __init__(self, column, name=None):
        """Set Attributes.

        Args:
            column (str): Column of the foreign key
            name (str, optional): Defaults to None. Name of the foreign key.
        """
        self._attributes = {
            'column': column,
            'name': name,
            'ref_table': None,
            'ref_column': None,
            'on_delete': None,
            'on_update': None
        }

    def references(self, ref_column):
        """Set the column the foreign key references.

        Args:
            ref_column (str): Name of referenced column

        Returns:
            ForeignKey: self instance for method chaining.
        """
        self._attributes['ref_column'] = ref_column
        return self

    def on(self, ref_table):
        """Set the table the foreign key references.

        Args:
            ref_table (str): Name of the referenced table

        Returns:
            ForeignKey: self instance for method chaining.
        """
        self._attributes['ref_table'] = ref_table
        return self

    def on_delete(self, action):
        """Set the action which will be executed on delete.

        Args:
            action (str): Action to execute

        Returns:
            ForeignKey: self instance for method chaining.
        """
        self._attributes['on_delete'] = action
        return self

    def on_update(self, action):
        """Set the action which will be executed on update.

        Args:
            action (str): Action to execute

        Returns:
            ForeignKey: self instance for method chaining.
        """
        self._attributes['on_update'] = action
        return self

    def get(self, attribute, default=None):
        """Get the value of an attribute, with a default if it does not exist.

        Args:
            attribute (str): Attribute you want to get
            default (object, optional): Defaults to None. Default value.

        Returns:
            object
        """
        return self._attributes.get(attribute, default)



"""Interface of the data store to make it possible to switch between different implementations."""
from __future__ import unicode_literals
from six import add_metaclass
from abc import abstractmethod, ABCMeta


@add_metaclass(ABCMeta)
class AbstractStore(object):

    """Contains the interface between the framework and the concrete Database implementation."""

    @abstractmethod
    def has_table(self, table):
        """Check if the data store has a specific table.

        Args:
            table (str): Name of the Table

        Raises:
            NotImplementedError: if method is not implemented by derived class.

        Returns:
            boolean
        """
        raise NotImplementedError('Must has_table rename method')

    def has_column(self, table, column):
        """Check if a table has a specific column.

        Args:
            table (str): Name of the table
            column (str): Name of the column

        Raises:
            NotImplementedError: if method is not implemented by derived class.

        Returns:
            boolean
        """
        raise NotImplementedError('Must has_column rename method')

    @abstractmethod
    def create(self, blueprint):
        """Create a new Table.

        Args:
            blueprint (Blueprint): The Blueprint to create the table

        Raises:
            NotImplementedError: if method is not implemented by derived class.
        """
        raise NotImplementedError('Must implement rename method')

    @abstractmethod
    def rename(self, old, new):
        """Rename a table.

        Args:
            old (str): Old table name
            new (str): New table name

        Raises:
            NotImplementedError: if method is not implemented by derived class.
        """
        raise NotImplementedError('Must implement rename method')

    @abstractmethod
    def drop(self, name, checkfirst=False):
        """Drop a table.

        Args:
            name (str): Name of the table

        Raises:
            NotImplementedError: if method is not implemented by derived class.
        """
        raise NotImplementedError('Must implement drop method')

    @abstractmethod
    def alter(self, blueprint):
        """Alter the schema for a table.

        Args:
            blueprint (Blueprint): Table Blueprint

        Raises:
            NotImplementedError: if method is not implemented by derived class.
        """
        raise NotImplementedError('Must implement alter method')



"""Helper functions for the webinterface."""
import re
import sys
from six import StringIO
from contextlib import contextmanager

ANSI_PATTERN = re.compile(r'\x1B\[[0-?]*[ -/]*[@-~]')


def ansi_escape(text):
    """Remove ansi colors from text.

    Args:
        text (str): Text to escape

    Returns:
        str: escaped text
    """
    return text
    # return ANSI_PATTERN.sub('', text)


@contextmanager
def capture_print(escape=False):
    """Capture prints to stdout and stderr.

    Args
        escape (bool, optional): Defaults to False. Whether to escape ansi sequences or not.
    """
    capturer = CapturedContent(escape)

    try:
        yield capturer
    finally:
        capturer.revert()


class CapturedContent(object):

    """Capture prints to stdout and stderr.

    Arguments:
        streams (dict): Output streams for stdout and stderr
        escape (bool): Flag to indicate if streams should be escaped.
    """

    def __init__(self, escape=True):
        """Set streams for stdout and stderr.

        Args:
            escape (bool, optional): Defaults to True. Whether to escape ansi sequences or not.
        """
        self.streams = {'out': StringIO(), 'err': StringIO()}
        self.escape = escape

        # capture stdout and stderr
        sys.stdout = self.streams['out']
        sys.stderr = self.streams['err']

    @classmethod
    def revert(cls):
        """Revert back to default stdout and stderr streams."""
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__

    def get_lines(self):
        """Get merged lines for stdout and stderr.

        Returns:
            list: list of lines
        """
        return self._merge()

    def get_text(self):
        """Get all prints/writes to stdout and stderr.

        Returns:
            str: text
        """
        return '\n'.join(self.get_lines())

    def clear(self):
        """Clear the streams."""
        self.streams['out'].seek(0)
        self.streams['err'].seek(0)

        self.streams['out'].truncate(0)
        self.streams['err'].truncate(0)

    def has_error(self):
        """Check if there is something in the error stream.

        Returns:
            boolean: Whether the error stream has content or not.
        """
        return self.streams['err'].getvalue() != ''

    def _prepare_lines(self, lines):
        """Strip and escape lines if needed.

        Args:
            lines (list): list of lines

        Returns:
            list: list of prepared lines
        """
        return list(map(lambda l: ansi_escape(l).strip() if self.escape else l.strip(), lines))

    def _merge(self):
        """Merge stdout and stderr together.

        Returns:
            list: list of lines
        """
        output = self._prepare_lines(self.streams['out'].getvalue().splitlines())
        output.extend(self._prepare_lines(self.streams['err'].getvalue().splitlines()))

        return output



"""Interface for using the *Repository Design Pattern*.

Instead of using the :py:class:`.AbstractStore` (or any of its implementations like
:py:class:`.Store`) as a *God Class* the *Repository Design Pattern*
allows you to follow the *Single Responsibility Rule*.

A Repository acts like an in-memory domain object collection, that connects the domain
and data mapping layers. Objects can be easily added to and removed from the Repository,
due to the mapping code of the Repository which will ensure that the right operations
are executed behind the scenes.

Apart from acting as a bridge between the data and the domain layer, the Repository
also provides some useful methods for storing, updating, deleting and querying data.

Examples
========
Lets assume that you are using the default experimentum data store :py:mod:`.Store`.
With :py:mod:`.Repository` there is already an implementation which will handle
the mapping between the Repository objects and the SQLAlchemy ORM which is used by
the :py:mod:`.Store`.

For simplicity we will assume that your database already contains a *User* and
*Address* table (for more information on how to specify your domain layer, see:
:py:mod:`.Migrations`).

Data Layer
----------
Each user has a ``name``, ``fullname`` and a ``password``
field and can have multiple ``adresses``. Each address has an ``email`` field::

    from experimentum.Storage import AbstractRepository


    class AddressRepository(AbstractRepository.implementation):
        __table__ = 'Address'

        def __init__(self, email):
            self.email = email


    class UserRepository(AbstractRepository.implementation):
        __table__ = 'User'

        __relationships__ = {
            'addresses': [AddressRepository]
        }

        def __init__(self, name, fullname, password):
            self.name = name
            self.fullname = fullname
            self.password = password

In the ``__init__`` method of the repository you specify the fields of the table.
The ``__table__`` attribute contains the name of the table, while the optional
``__relationships__`` attribute specifies any relationships of the data.
The key is the attribute under which the data will be accessible, ie::

    print(user.addresses)  # will print all addresses of a specific user

Actions
-------
There are several methods for querying data, like :py:meth:`~.Repository.find`,
:py:meth:`~.Repository.get`, :py:meth:`~.Repository.first`, :py:meth:`~.Repository.all`::

    # Find a user by its id
    user = UserRepository.find(2)  # get user with ID 2
    print(user.id, user.name, user.fullname, user.addresses)

    # Get many users by a condition (here where name != 'John')
    users = UserRepository.get(where=['name', '!=', 'John'])
    for user in users:
        print(user.id, user.name, user.fullname, user.addresses)

    # Get the first user which satisfies a certain condition
    user = UserRepository.first(where=['name', 'John'])
    print(user.id, user.name, user.fullname, user.addresses)

    # Get all users
    users = UserRepository.all()
    for user in users:
        print(user.id, user.name, user.fullname, user.addresses)

The :py:class:`.Repository` class also provides several methods for storing, updating, and
deleting data, like :py:meth:`~.Repository.create`, :py:meth:`~.Repository.update`
:py:meth:`~.Repository.delete`::

    # Create a new user from a data dictionary
    user = UserRepository.from_dict({
        'name': 'Hello',
        'fullname': 'World',
        'password': '1234',
        # Turns the entries into AddressRepository instances and adds them to the user repo
        # so that they get saved with the correct user id
        'addresses': [
            {'email': 'hello@world.com'},
            {'email': 'foo@world.com'},
        ]
    })
    user.create()  # create the new user

    # Update a user
    user = UserRepository.find(2)
    user.fullname = 'Doe'
    user.name = 'John'
    user.update()

    # Delete a user
    user = UserRepository.find(1)
    user.delete()

Events
------
A Repository provides several events, allowing you to hook into the following points in a
repositorie's lifecycle: :py:meth:`~.AbstractRepository.before_insert`,
:py:meth:`~.AbstractRepository.after_insert`, :py:meth:`~.AbstractRepository.before_update`,
:py:meth:`~.AbstractRepository.after_update`.

Events allow you to easily execute code each time after/before a specific repository object
is saved or updated in the data store.

::

    class AddressRepository(AbstractRepository.implementation):
        # ...

        def after_insert(self):
            print('Gets called each time after an address is saved to the database.')
            print(self.id, self.email)  # Object has access to the inserted id

Loading
-------
In order for the framework to map all the repositories it has to load them via the
:py:class:`~.RepositoryLoader` class. For the :py:class:`~.RepositoryLoader` class
to be able to find the repositories, the files have be in your repository folder and
end with ``Repository.py``. The repository files should contain a repository class with
the same name as the file name.
"""
import os
import sys
import json
from six import add_metaclass
from abc import abstractmethod, ABCMeta
from experimentum.cli import print_failure
from experimentum.utils import find_files


class RepositoryLoader(object):

    """Load and map all the repositories it can find and cache them.

    Attributes:
        app (App): Main App class.
        implementation (AbstractRepository): Concrete repo implementation which should be used.
        store (AbstractStore): Data store which is used
    """

    def __init__(self, app, implementation, store):
        """Set up loader.

        Args:
            app (App): Main App class.
            implementation (AbstractRepository): Concrete repo implementation which should be used.
            store (AbstractStore): Data store which is used
        """
        self.app = app
        self.implementation = implementation
        self.store = store
        self._repos = {}

        AbstractRepository.implementation = implementation

    def _import_file(self, name, loc):
        """Load the repository file modules.

        Args:
            name (str): Name of the Repository
            loc (str): Location of the Repository

        Returns:
            Module: Repository Module

        Raises:
            Error: if the module could not be found
        """
        path = self.app.config.get('storage.repositories.path', 'repositories')

        if sys.version_info >= (3, 5):
            import importlib.util
            import importlib.machinery

            # finder is needed when folder is not a subfolder of root dir (e.g. for test layouts)
            loader_details = (
                importlib.machinery.SourceFileLoader,
                importlib.machinery.SOURCE_SUFFIXES
            )

            finder = importlib.machinery.FileFinder(loc, loader_details)
            spec = finder.find_spec(name)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            return module

        else:
            import imp
            # first find and load the package assuming it is in
            # the current working directory, '.'
            handler, pathname, desc = imp.find_module(os.path.join(self.app.root, path), ['.'])
            pkg = imp.load_module(path, handler, pathname, desc)

            # then find the named repository module using pkg.__path__
            # and load the module using the dotted name
            handler, pathname, desc = imp.find_module(name, pkg.__path__)
            module = imp.load_module(path + '.' + name, handler, pathname, desc)
            handler.close()

            return module

    def load(self):
        """Load all repositories from the repo folder and try to map them to the store."""
        # Get Repository files
        files = find_files(
            self.app.root,
            self.app.config.get('storage.repositories.path', 'repositories')
        )

        for filename in files:
            name = os.path.basename(filename)[:-3]

            # Try to load and map the repos
            try:
                module = self._import_file(name, os.path.abspath(os.path.dirname(filename)))
                repo = getattr(module, name)
                repo.mapping(repo, self.store)
                self._repos[name] = repo

            except Exception as exc:
                print_failure('Could not load repository. ' + str(exc), exit_code=1)

    def get(self, repository):
        """Return class of a loaded repository if it exists.

        Args:
            repository (str): Name of repository class.

        Returns:
            AbstractRepository: Repository class
        """
        repo = self._repos.get(repository, None)

        if repo is None:
            print_failure(
                'Could not load repository {}. Are you sure it exists?'.format(repository),
                exit_code=1
            )

        return repo


@add_metaclass(ABCMeta)
class AbstractRepository(object):

    """Interface for using the *Repository Design Pattern*.

    Instead of using the :py:class:`.AbstractStore` as a *God Class* the
    *Repository Design Pattern*  allows you to follow the *Single Responsibility Rule*.

    Attributes:
        store (AbstractStore): Store that is used for mapping domain and data layer.
        implementation(AbstractRepository): Concrete repo implementation which should be used.
        __table__ (str): Name of the table the repository refers to.
        __relationship__ (dict): Any Relationships the data has.
    """

    implemantation = None
    store = None
    __table__ = ''
    __relationships__ = {}

    def __init__(self, **attributes):
        """Set all attributes which where passed as kwargs."""
        for attr, val in attributes.items():
            self[attr] = val

    def to_json(self):
        """Return JSON representation of repository fields and values.

        Returns:
            str: JSON representration
        """
        attrs = {k: v for k, v in self.__dict__.items() if k[:1] != '_'}
        return json.dumps(attrs, sort_keys=True, indent=4)

    def __repr__(self):
        """Human readable representation of repository fields and values.

        Returns:
            str: human readable representation
        """
        attrs = ['{}="{}"'.format(k, v) for k, v in self.__dict__.items() if k[:1] != '_']
        return '<{} {} />'.format(self.__class__.__name__, ' '.join(attrs))

    def __contains__(self, key):
        """Check if repository has a certain attribute.

        Args:
            key (str): Attribute to check

        Returns:
            bool: Whether the attribute exists or not
        """
        return hasattr(self, key)

    def __getitem__(self, key):
        """Get attribute value with dictionary like syntax.

        Args:
            key (str): Attribute Key

        Returns:
            object: Attribute value
        """
        return getattr(self, key)

    def __setitem__(self, key, item):
        """Set attribute value with dictionary like syntax.

        Args:
            key (str): Attribute Key
            item (object): Attribute value
        """
        setattr(self, key, item)

    @classmethod
    def from_dict(cls, data):
        """Create a new Repository instance based on a dictionary entry.

        Args:
            data (dict): Repository Data

        Returns:
            AbstractRepository: Repository instance filled with data.
        """
        init = {}
        relations = {}
        relationships = cls.__relationships__

        for key, val in data.items():
            if key in relationships:
                if isinstance(val, list):
                    relations[key] = [relationships[key][0].from_dict(v) for v in val]
                else:
                    relations[key] = [relationships[key][0].from_dict(val)]
            else:
                init[key] = val

        try:
            repo = cls(**init)
        except TypeError as err:
            # Unexpected Argument
            if 'unexpected keyword argument' in str(err):
                print_failure(
                    "{}. Either fix the typo or add it the repository constructor.".format(
                        str(err).replace('__init__()', cls.__name__),
                    ),
                    1
                )
            # Determine missing parameters
            elif 'required positional argument' in str(err) or 'takes exactly' in str(err):
                import inspect
                if hasattr(inspect, 'getfullargspec'):
                    getargspec = inspect.getfullargspec
                else:
                    getargspec = inspect.getargspec

                available = set(getargspec(cls.__init__).args[1:])
                passed = set(init.keys())
                print_failure(
                    "The {} class is missing the following parameters: {}".format(
                        cls.__name__, available - passed
                    ),
                    2
                )
            else:
                print_failure(err, -1)

        for key, val in relations.items():
            repo[key] = val

        return repo

    @staticmethod
    def mapping(cls, store):
        """Map data store content to repository classes.

        Example:
        Map a *User* Table in the data store to the *UserRepository*,
        which reflects the attributes of the User Table schema.

        Args:
            cls (AbstractRepository): Repository to map
            store (AbstractStore): Storage that is used

        Raises:
            NotImplementedError: if method is not implemented
        """
        raise NotImplementedError('Must implement mapping method!')

    @abstractmethod
    def create(self):
        """Save the repository content in your data store.

        Raises:
            NotImplementedError: if method is not implemented yet.
        """
        raise NotImplementedError('Must implement create method!')

    @abstractmethod
    def update(self):
        """Update the repository content in your data store.

        Raises:
            NotImplementedError: if method is not implemented yet.
        """
        raise NotImplementedError('Must implement update method!')

    @abstractmethod
    def delete(self):
        """Delete the repository content from your data store.

        Raises:
            NotImplementedError: if method is not implemented yet.
        """
        raise NotImplementedError('Must implement delete method!')

    @classmethod
    def get(cls, where=None):
        """Get all entries which satisfy a specific condition from your data store.

        Args:
            where (list, optional): Defaults to None. Where Condition

        Raises:
            NotImplementedError: if method is not implemented yet.

        Returns:
            list: List of items which satisfy the condition.
        """
        raise NotImplementedError('Must implement get method!')

    @classmethod
    def first(cls, where=None):
        """Get first entry which satisfy a specific condition from your data store.

        Args:
            where (list, optional): Defaults to None. Where Condition

        Raises:
            NotImplementedError: if method is not implemented yet.

        Returns:
            AbstractRepository: Item which satisfies the condition.
        """
        raise NotImplementedError('Must implement first method!')

    @classmethod
    def all(cls):
        """Get all entries for this specific repository from your data store.

        Raises:
            NotImplementedError: if method is not implemented yet.

        Returns:
            list: List of all entires
        """
        raise NotImplementedError('Must implement all method!')

    @classmethod
    def find(cls, id):
        """Find an entry of this repository based on its id.

        Args:
            id (int): ID to search for.

        Raises:
            NotImplementedError: if method is not implemented yet.

        Returns:
            AbstractRepository: Item which the concrete id
        """
        raise NotImplementedError('Must implement find method!')

    def before_insert(self):
        """Event that gets called before insert statement is executed.

        Returns:
            AbstractRepository
        """
        return self

    def after_insert(self):
        """Event that gets called after insert statement is executed.

        Returns:
            AbstractRepository
        """
        return self

    def before_update(self):
        """Event that gets called before update statement is executed.

        Returns:
            AbstractRepository
        """
        return self

    def after_update(self):
        """Event that gets called after update statement is executed.

        Returns:
            AbstractRepository
        """
        return self



"""Some Helper functions for creating files and folders."""
import os
import json


def create_folder(root, path, with_init=False):
    """Create a folder if it does not exist.

    Args:
        root (str): Root folder
        path (str): folder path
        with_init (bool): Add an __init__.py to folder or not
    """
    path = os.path.realpath(os.path.join(root, path))
    if os.path.exists(path) is False:
        os.makedirs(path)

    if with_init:
        with open(os.path.join(path, '__init__.py'), 'w+') as filehandler:
            filehandler.write('')


def create_config_file(root, path, name, data):
    """Create a config file and fill it with data.

    Args:
        root (str): Root folder
        path (str): Path to config folder
        name (str): Name of the config file
        data (object): Config Data
    """
    path = os.path.realpath(os.path.join(root, path))
    with open(os.path.join(path, name), 'w+') as cfg:
        json.dump(data, cfg, sort_keys=False, indent=4, separators=(',', ': '))


def create_from_stub(stubfile, filename, attrs):
    """Create a new file from a stub.

    Args:
        stubfile (str): Name of stub file in _stubs folder
        filename (str): Filename to create (with path)
        attrs (dict): Attributes to pass to stub
    """
    # Load Stub file
    stub = ''
    with open(os.path.join(os.path.dirname(__file__), '_stubs', stubfile), 'r') as filehandler:
        stub = filehandler.read()

    # Create new file based on stub
    with open(filename, 'w+') as filehandler:
        filehandler.write(stub.format(**attrs))



"""SQLite specific sql commands and queries.

SQLite does not provide complete range of all sql commands, therefore some tricks are needed
to emulate the behavior. See: https://www.sqlite.org/omitted.html
"""
import logging
from experimentum.Storage.SQLAlchemy import Platform
from sqlalchemy import Table, Index, ForeignKey


def prepare_columns(columns, indexes, dropped):
    """Prepate columns so that they can be created for another table.

    Args:
        columns (list): Columns of a table
        indexes (list): indexes of a table
        dropped (object): Columns and indexes to drop

    Returns:
        dict: with columns and column names
    """
    data = {'columns': [], 'names': []}
    dropped_fkeys = [fkey['name'] for fkey in dropped['indexes'] if fkey['type'] == 'foreign']
    dropped_pkeys = [pkey['col'] for pkey in dropped['indexes'] if pkey['type'] == 'primary']

    # Add Columns
    for col in columns:
        column = col.copy()

        # Add Foreign Keys
        fkeys = []
        for fkey in col.foreign_keys:
            if fkey.name not in dropped_fkeys:
                fkeys.append(ForeignKey(
                    fkey.column,
                    name=fkey.name,
                    onupdate=fkey.onupdate,
                    ondelete=fkey.ondelete,
                    deferrable=fkey.deferrable
                ))
        column.foreign_keys = set(fkeys)

        # Drop primary keys
        if column.primary_key and column.name in dropped_pkeys:
            column.primary_key = False

        # Add column and names
        if column.name not in dropped['columns']:
            data['columns'].append(column)
            data['names'].append(column.name)

    # Add Indexes
    for idx in indexes:
        drop = False
        for drop_idx in dropped['indexes']:
            if drop_idx['name'] == idx.name:
                drop = True
                break

        if drop is False:
            data['columns'].append(idx)

    return data


class SQLitePlatform(Platform):

    """SQLite specific sql commands and queries."""

    def alter_table(self, table, cols, dropped):
        """Alter SQLite Table.

        SQLite does not provide complete ALTER TABLE support, only RENAME TABLE and ADD COLUMN
        variants are supported. In order to add columns and keys we have to create a tempory
        table which holds the data of the table and then drop and recreate the table with
        the modified schema. Lastly we have to copy the data back into the new table and
        delete the temporary data table.

        See:
        https://www.sqlite.org/omitted.html

        Args:
            table (str): Name of the table
            cols (list): List with changed columns
            dropped (dict): Dictionary with list of columns and indexes to drop
        """
        # Prepare columns
        old_table = self.meta.tables.get(table, Table(table, self.meta, autoload=True))
        columns = prepare_columns(list(old_table.columns), list(old_table.indexes), dropped)
        new_columns = columns['columns']
        new_columns.extend(cols)

        # * CREATE TABLE $data_table AS SELECT ($old_cols) FROM $old_table;
        data_table = self._create_table('__tmp__{}'.format(table), columns['columns'])
        self._fill_table(data_table, columns['names'], old_table)

        # * DROP TABLE $old_table;
        old_table.drop(self.engine)

        # * CREATE TABLE $new_table ($old_columns + $new_columns);
        self._create_table('__new__{}__'.format(table), new_columns)
        self.engine.execute(self.get_rename_sql('__new__{}__'.format(table), table))

        # * INSERT INTO $new_table ($old_column) SELECT * FROM $data_table;
        self._fill_table(table, columns['names'], data_table)

        # * DROP TABLE $data_table;
        data_table.drop(self.engine)

    def _fill_table(self, target_tablename, columns, source_table):
        """Fill a target table with values from a source table.

        Args:
            target_tablename (str): Table to be filled
            columns (list): List of columns to be filled
            source_table (str): Name of the source table
        """
        self.engine.execute('INSERT INTO {} ({}) SELECT {} FROM {}'.format(
            target_tablename,
            ','.join(columns),
            ','.join(columns),
            source_table
        ))

    def _create_table(self, tablename, columns):
        """Create a new table with columns and constraints.

        Args:
            tablename (str): Name of new table
            columns (list): List of Columns

        Returns:
            Table: newly created table
        """
        table = Table(tablename, self.meta, extend_existing=True)
        used_idx = []

        for column in columns:
            if isinstance(column, Index):
                # copy columns to avoid association with new table errors
                idx_cols = [idx_col.copy() for idx_col in column.columns]

                # Add index
                if column.name not in used_idx:
                    table.append_constraint(Index(column.name, *idx_cols, unique=column.unique))
                    used_idx.append(column.name)

                # Drop old index if exists to avoid naming conflicts
                try:
                    column.drop(self.engine)
                except Exception as exc:
                    # do not fail, just keep going - but log the error (just in case)
                    logging.getLogger('experimentum').warning(
                        'Error while recreating index: ' + str(exc)
                    )
            else:
                table.append_column(column.copy())

        table.create(self.engine, checkfirst=True)
        return table



"""Column factory class to create new columns.

Used to create new columns, indexes and foreign keys
with the help of sqlalchemy.
"""
from sqlalchemy import Column, Index, ForeignKey
from sqlalchemy.dialects.mysql import INTEGER, BIGINT, DOUBLE, LONGTEXT, MEDIUMINT, MEDIUMTEXT
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.types import ARRAY, BigInteger, Integer, Boolean, Date, DateTime, Enum,\
    LargeBinary, Numeric, SmallInteger, String, Text, Time, CHAR, Float, JSON, TIMESTAMP


def get_number_type(col_type, params, unsigned=False):
    """Create a number type column.

    Args:
        col_type (string): Type of the column.
        params (object): Additional parameters.
        unsigned (bool, optional): Defaults to False. Whether or not it is an unsigned int.

    Returns:
        sqlalchemy.types.TypeEngine: Number type like integer or float
    """
    # Todo: Unsinged Integers, check if they work in mysql, postgresql etc
    unsigned = True if col_type == 'big_increments' or col_type == 'increments' else unsigned

    # Integers
    if col_type == 'big_increments' or col_type == 'big_integer':
        return BigInteger()\
            .with_variant(BIGINT(unsigned=unsigned), 'mysql')\
            .with_variant(Integer(), 'sqlite')
    elif col_type == 'increments' or col_type == 'integer':
        return Integer().with_variant(INTEGER(unsigned=unsigned), 'mysql')
    # Floats
    elif col_type == 'float':
        # ! Seems not to work, don't know why???
        return Float(params.get('precision'), decimal_return_scale=params.get('scale'))
    elif col_type == 'decimal':
        return Numeric(precision=params.get('precision'), scale=params.get('scale'))
    elif col_type == 'double':
        return DOUBLE(
            precision=params.get('precision'),
            scale=params.get('scale')
        ).with_variant(
            Float(params.get('precision'), decimal_return_scale=params.get('scale')), 'sqlite'
        )
    elif col_type == 'medium_integer':
        return MEDIUMINT().with_variant(INTEGER(unsigned=unsigned), 'sqlite')
    elif col_type == 'small_integer':
        return SmallInteger()


def get_string_type(col_type, params):
    """Create a string type column.

    Args:
        col_type (string): Type of the column.
        params (object): Additional parameters.

    Returns:
        sqlalchemy.types.TypeEngine: String type like char or text
    """
    if col_type == 'char':
        return CHAR(params.get('length'))
    elif col_type == 'json':
        return (
            JSON(none_as_null=True)
            .with_variant(JSONB(none_as_null=True), 'postgresql')
            .with_variant(Text(), 'sqlite')
        )
    elif col_type == 'long_text':
        return LONGTEXT().with_variant(Text(), 'sqlite')
    elif col_type == 'medium_text':
        return MEDIUMTEXT().with_variant(Text(), 'sqlite')
    elif col_type == 'string':
        return String(length=params.get('length'))
    elif col_type == 'text':
        return Text()


def get_time_type(col_type):
    """Create a time type column.

    Args:
        col_type (string): Type of the column.

    Returns:
        sqlalchemy.types.TypeEngine: Time type like date or timestamp
    """
    if col_type == 'date':
        return Date()
    elif col_type == 'datetime':
        return DateTime()
    elif col_type == 'time':
        return Time()
    elif col_type == 'timestamp':
        return TIMESTAMP()


class ColumnFactory(object):

    """Factory class to create SQLAlchemy Columns."""

    @staticmethod
    def get_indexes(col, indexes, used_indexes=None):
        """Get all indexes and primary keys of a column.

        Args:
            col (Column): The column
            indexes (list): List of indexes
            used_indexes (list, optional): Defaults to None. Used indexes cols for composite keys.

        Returns:
            tuple: with primary keys, indexes and used indexes
        """
        used_indexes = [] if used_indexes is None else used_indexes
        _indexes = []
        primary_key = False

        for idx in indexes:
            idx_type = idx.get('type')
            idx_cols = idx.get('col')

            if idx_cols == col.get('name') or col.get('name') in idx_cols:
                # Primary Key
                if idx_type == 'primary':
                    primary_key = True
                # Indexes
                elif idx_type == 'unique' or idx_type == 'index' and idx_cols not in used_indexes:
                    if not isinstance(idx_cols, list):
                        idx_cols = [idx_cols]

                    used_indexes.extend(idx_cols)
                    _indexes.append(
                        Index(idx.get('name'), *idx_cols, unique=idx_type == 'unique')
                    )

        return primary_key, _indexes, used_indexes

    def get_columns_and_indexes(self, blueprint):
        """Get Column and Index instances from the blueprint.

        Args:
            blueprint (Blueprint): Schema Blueprint to add/modify table.

        Returns:
            dict: dictionary with columns and indexes
        """
        # Collection of columns and indexes
        used_indexes = []
        data = {
            'columns': [],
            'indexes': [],
        }

        for col in blueprint.columns:
            # Get Indexes and Primary Keys
            primary, indexes, _used_idxs = self.get_indexes(col, blueprint.indexes, used_indexes)
            data['indexes'].extend(indexes)
            used_indexes.extend(_used_idxs)

            # Add Foreign Key column
            foreign_key = self.get_foreign_key(col, blueprint.fkeys)
            if foreign_key is not False:
                data['columns'].append(foreign_key)
                continue

            # Add Column
            data['columns'].append(Column(
                col.get('name'),
                self.get_type(
                    col.get('type'),
                    col.get('parameters'),
                    unsigned=col.get('unsigned')
                ),
                default=col.get('default'),
                nullable=col.get('null'),
                autoincrement=col.get('parameters', {}).get('autoincrement', 'auto'),
                primary_key=primary,
            ))

        return data

    def create_foreign_key(self, key, col):
        """Create a foreign key column.

        Arguments:
            key (ForeignKey): Foreign Key
            col (Column): column to create foreign key on

        Returns:
            sqlalchemy.schema.Column: foreign key column
        """
        fkey = ForeignKey(
            key.get('ref_table') + '.' + key.get('ref_column'),
            ondelete=key.get('on_delete'),
            onupdate=key.get('on_update'),
            name=key.get('name')
        )

        return Column(
            col.get('name'),
            self.get_type(
                col.get('type'),
                col.get('parameters'),
                unsigned=col.get('unsigned')
            ),
            fkey,
            default=col.get('default'),
            nullable=col.get('null'),
            autoincrement=col.get('autoincrement', False)
        )

    def get_foreign_key(self, col, fkeys):
        """Get a foreign key from a column.

        Args:
            col (Column): Column to get foreign key of.
            fkeys (list): List of foreign keys

        Returns:
            sqlalchemy.schema.Column, False: either a foreign key column if found or False
        """
        col_name = col.get('name')
        foreign_key = [
            self.create_foreign_key(fkey.get('key'), col)
            for fkey in fkeys
            if fkey.get('col') == col_name  # or col_name in fkey.get('col')
        ]

        if len(foreign_key):
            return foreign_key[0]

        return False

    def get_type(self, col_type, params=None, unsigned=False):
        """Map the type to valid Column Types.

        Notes:
            http://docs.sqlalchemy.org/en/latest/core/type_basics.html

        Args:
            col_type (str): Type of column
            params (dict, optional): Defaults to None. Additional Column Options.
            unsigned (bool, optional): Defaults to False. If it is an unsigned integer or not.

        Returns:
            sqlalchemy.types.TypeEngine: Type for new column
        """
        # TODO: Check if vendor specific types like json, mediumint, etc work
        params = {} if params is None else params

        # Get number types
        if (
            'integer' in col_type or 'increments' in col_type or col_type == 'decimal' or
            col_type == 'double' or col_type == 'float'
        ):
            return get_number_type(col_type, params, unsigned)
        # Get String types
        elif (
            'text' in col_type or col_type == 'char' or col_type == 'json' or col_type == 'string'
        ):
            return get_string_type(col_type, params)
        # Get Date/Time Types
        elif 'date' in col_type or 'time' in col_type:
            return get_time_type(col_type)
        # Get BINARY type
        elif col_type == 'binary':
            return LargeBinary()
        # Get Boolean type
        elif col_type == 'boolean':
            return Boolean()
        # Get Enum Type
        elif col_type == 'enum':
            return Enum(*params.get('fields', []))
        # Get Array type
        elif col_type == 'array':
            arr_type = self.get_type(params.get('arr_type', 'text'))
            return arr_type.with_variant(
                ARRAY(arr_type, dimensions=params.get('dimensions')),
                'postgresql'
            )

        return Text()



"""Contains all views/routes for the plots.

Available Routes for:
    * Generating plots of an experiment
    * Exporting plots of an experiment
    * Serving plot image files
"""
from flask import Blueprint, current_app, send_from_directory, jsonify, url_for, send_file,\
    abort, redirect
from werkzeug.utils import secure_filename
import os
import shutil
import tempfile


blueprint = Blueprint('plots', __name__)


def generate_plots(directory, container, experiment):
    """Generate all plots for an experiment.

    Args:
        directory (str): Plot Upload directory.
        container (App): Service Container.
        experiment (str): Name of the experiment.

    Returns:
        list: list of status messages.
    """
    # Find plots for experiment and generate them
    messages = []
    config = container.config.get('plots', {})

    # create experiment plot dir if not exists
    if not os.path.exists(directory):
        os.makedirs(directory)

    for name, cfg in config.items():
        if cfg.get('experiment') == experiment:
            # try to plot and save as svg
            try:
                plot = container.make('plot', name)
                plot.plotting()
                plot.save(os.path.join(directory, name + '.svg'))

                messages.append({
                    'message': 'Generated plot: ' + name,
                    'status': 'success',
                    'file': url_for('plots.image', experiment=experiment, filename=name + '.svg')
                })
            except Exception as exc:
                messages.append({
                    'message': 'Error generating plot {}: {}'.format(name, exc),
                    'status': 'error',
                    'file': None
                })

    return messages


@blueprint.route('/export/<experiment>')
def export(experiment):
    """Download all plots for an experiment.

    Args:
        experiment (str): Name of the experiment.

    Returns:
        Response: File Download
    """
    directory = os.path.join(current_app.config['UPLOAD_FOLDER'], experiment)
    tmpfile = os.path.join(tempfile.gettempdir(), 'export-' + experiment)

    try:
        shutil.make_archive(tmpfile, 'zip', directory)
        return send_file(tmpfile + '.zip', as_attachment=True)
    except Exception:
        current_app.config.get('container').log.error('No such file or directory: %s' % directory)
        abort(400)


@blueprint.route('/generate/<experiment>')
def generate(experiment):
    """Generate all plots for an experiment.

    Args:
        experiment (str): Name of the experiment.

    Returns:
        Response: Redirects to plot overview.
    """
    # Find plots for experiment and generate them
    container = current_app.config.get('container')
    directory = os.path.join(current_app.config['UPLOAD_FOLDER'], experiment)
    generate_plots(directory, container, experiment)
    return redirect(url_for('experiments.plots', experiment=experiment))


@blueprint.route('/generate_ajax/<experiment>')
def generate_ajax(experiment):
    """Genereate all plots for an experiment.

    Args:
        experiment (str): Name of the experiment.

    Returns:
        str: JSON with status messages
    """
    # Find plots for experiment and generate them
    container = current_app.config.get('container')
    directory = os.path.join(current_app.config['UPLOAD_FOLDER'], experiment)
    messages = generate_plots(directory, container, experiment)
    return jsonify(messages)


@blueprint.route('/image/<experiment>/<filename>')
def image(experiment, filename):
    """Serve plot images from the plot graphics directory.

    Args:
        experiment (str): Name of the experiment.
        filename (str): Name of the plot.

    Returns:
        obj: Plot graphics
    """
    return send_from_directory(
        os.path.join(current_app.config['UPLOAD_FOLDER'], experiment),
        secure_filename(filename)
    )



"""Import classes for easier importing by other packages/modules."""
# flake8: noqa
from .AbstractCommand import command, AbstractCommand
from .CommandManager import CommandManager
from .MigrationCommand import status, refresh, up, down, make
from .ExperimentsCommand import run
from .PlotCommand import generate
from .WebGUICommand import start



"""Contains all views/routes for the dashboard.

The dashboard lists:
* all available experiments
* all migrations and their status

  * Upgrade migrations
  * Downgrade migrations
  * Refresh Migrations
"""
from experimentum.WebGUI.views.migrations import get_migration_status
from experimentum.Experiments import Experiment
from flask import Blueprint, current_app, render_template
from sqlalchemy.exc import InvalidRequestError

blueprint = Blueprint('dashboard', __name__)


@blueprint.route('/')
def dashboard():
    """Displays the dashboard template.

    Returns:
        str: HTML Template
    """
    data = {}

    try:
        data = {
            'experiments': Experiment.get_status(current_app.config.get('container')),
            'migrations': get_migration_status(),
        }
    except InvalidRequestError as exc:
        current_app.config.get('container').log.critical(exc)
        data = {
            'error': 'There seems to be an error with your database. \
            Please try to refresh your migrations and restart the webgui to resolve this problem.',
            'migrations': get_migration_status(),
        }

    return render_template('dashboard/index.jinja', **data)



"""The config module stores all config items.

The Config class supports the *dot-notation* which means that you
can access nested elements seperated by a ``.``, e.g. ``foo.bar.baz``.

Example:

.. code-block:: python

    cfg = Config()
    cfg.set({ 'foo': 'bar', 'foobar': {'baz': 42 } })
    cfg.set('a.b', 'c')

    cfg.get('foobar.baz')  # returns: 42
    cfg.has('foobar.foo')  # returns: False
    cfg.all()              # returns: { 'foo': 'bar', 'foobar': {'baz': 42 }, 'a': {'b': 'c'} }

"""


class Config(object):

    """Manages all the config items.

    Attributes:
        items (dict): Loaded config items.
    """

    def __init__(self):
        """Init config store."""
        self.items = {}

    def has(self, key):
        """Determine if the given configuration value exists.

        Args:
            key (string): config item key

        Returns:
            bool
        """
        return True if self._dot(key, False) is not False else False

    def get(self, key, default=None):
        """Get the specified configuration value.

        Args:
            key (string|list): Set of keys.
            default (object, optional): Defaults to None. A default value if the key is
                not found

        Returns:
            object: The config value.
        """
        if isinstance(key, list):
            return self.get_many(key)

        return self._dot(key, default)

    def all(self):
        """Get all of the configuration items for the application.

        Returns:
            dict: config items
        """
        return self.items

    def get_many(self, keys):
        """Get many configuration values.

        Arguments:
            keys (list): Names of the keys.

        Returns:
            dict: Donfig items for the set of keys.
        """
        return {
            key: self._dot(key, default) for key, default in keys
        }

    def _dot(self, path, default):
        """Get an item using using "dot" notation.

        Arguments:
            path (string): Path in dot notation
            default (object): Default value

        Returns:
            object: config item
        """
        arr = self.items
        keys = path.split('.')

        for key in keys:
            arr = arr.get(key, default)
            if not isinstance(arr, dict):
                return arr

        return arr

    def set(self, key, value=None):
        """Set a given configuration value.

        Arguments:
            key (dict|string): Config item keys to change the value
            value (object,optional): Defaults to None. New value
        """
        def set_to(items, data):
            """Recursively set value to dictionary deep key.

            Args:
                items (list): List of dictionary keys
                data (dict): Portion of dictionary to operate on
            """
            item = items.pop(0)

            # Traverse dictionary to last item key
            if items:
                if item not in data:
                    data[item] = {}
                set_to(items, data[item])
            # Set Value on last key
            else:
                data[item] = value

        if isinstance(key, str):
            set_to(key.split('.'), self.items)
        elif isinstance(key, dict):
            self.items.update(key)



"""Handle all migration related stuff, and provides the following classes/modules.

:py:mod:`.Schema`
    The :py:class:`.Schema` Builder lets you create, delete, or alter tables.

:py:mod:`.Blueprint` Module
    The :py:class:`.Blueprint` class lets you create, delete, or alter columns of a table.

:py:mod:`.Migration` Module
    The :py:class:`.Migration` class handles database migrations.

:py:mod:`.Migrator` Module
    The :py:class:`.Migrator` class handles the management of all migrations.

:py:mod:`.Column` Module
    Datastructure for columns.

:py:mod:`.ForeignKey` Module
    Datastructure for foreign keys.
"""
# flake8: noqa
from .Migrator import Migrator
from .Migration import Migration
from .Column import Column
from .ForeignKey import ForeignKey
from .Blueprint import Blueprint
from .Schema import Schema



"""Plot CLI commands to allow you to generate your plots and charts.

Generating plots and charts
---------------------------
Use the ``plots:generate`` command to generate a plot/chart.

Arguments:

====  ==========================================================================
name  Name of the plot/chart *(plot|chart|graph is omitted from the filename)*.
====  ==========================================================================

Options:

-o          Output file where the plot is stored at.
            When omitted the plot is shown directly.
-h, --help  Show the help message.
"""
from experimentum.Commands import command


@command('', help='Generate a plot', arguments={
    'name': {
        'type': str, 'help': 'Name of the plot.'
    },
    '-o': {
        'help': 'Output', 'type': str
    }
})
def generate(app, args):
    """Generate a new plot.

    Args:
        app (App): App Service Container.
        args (argparse.Namespace): Command Arguments and Options.
    """
    plot = app.make('plot', args.name)
    plot.plotting()

    if hasattr(args, 'o') and args.o:
        plot.save(args.o)
    else:
        plot.show()



r"""The :py:class:`.Blueprint` class lets you create, delete, or alter columns of a table.

Columns
=======

Adding Columns
--------------
To update an existing table, you can use the :py:meth:`.Schema.table` method.
The method accepts a table name as its argument and returns a :py:class:`.Blueprint`
instance that can be used to add columns to the table::

    with self.schema.table('users') as table:
        table.string('email')
        table.string('address').nullable()
        table.integer('age').nullable().unsigned()

The table builder contains a plethora of column types that you can use when building your tables:

+-------------------------------------------+------------------------------------------------------+
| Command                                   | Description                                          |
+===========================================+======================================================+
| ``table.array('ratings', 'integer', 2)``  | ARRAY column type. Only supported in **postgresql**! |
+-------------------------------------------+------------------------------------------------------+
| ``table.big_increments('id')``            | Auto-incrementing ID using a "big integer" (primary  |
|                                           | key) equivalent column.                              |
+-------------------------------------------+------------------------------------------------------+
| ``table.big_integer('votes)``             | BIGINT equivalent column.                            |
+-------------------------------------------+------------------------------------------------------+
| ``table.binary('data')``                  | BLOB equivalent column.                              |
+-------------------------------------------+------------------------------------------------------+
| ``table.boolean('confirmed')``            | BOOLEAN equivalent column.                           |
+-------------------------------------------+------------------------------------------------------+
| ``table.char('name', 4)``                 | CHAR equivalent column with a length.                |
+-------------------------------------------+------------------------------------------------------+
| ``table.date('created_on')``              | DATE equivalent column.                              |
+-------------------------------------------+------------------------------------------------------+
| ``table.datetime('created_at')``          | DATETIME equivalent column.                          |
+-------------------------------------------+------------------------------------------------------+
| ``table.decimal('amount', 5, 2)``         | DECIMAL equivalent column with a precision           |
|                                           | *(total digits)* and a scale *(decimal digits)*.     |
+-------------------------------------------+------------------------------------------------------+
| ``table.double('column', 15, 8)``         | DOUBLE equivalent column with a precision            |
|                                           | *(total digits)* and a scale *(decimal digits)*.     |
+-------------------------------------------+------------------------------------------------------+
| ``table.enum('choices', ['foo', 'bar'])`` | ENUM equivalent column.                              |
+-------------------------------------------+------------------------------------------------------+
| ``table.float('amount')``                 | FLOAT equivalent column.                             |
+-------------------------------------------+------------------------------------------------------+
| ``table.increments('id')``                | Auto-incrementing ID using a "integer" (primary key) |
|                                           | equivalent column.                                   |
+-------------------------------------------+------------------------------------------------------+
| ``table.integer('votes')``                | INTEGER equivalent column.                           |
+-------------------------------------------+------------------------------------------------------+
| ``table.json('options')``                 | JSON equivalent column.                              |
+-------------------------------------------+------------------------------------------------------+
| ``table.long_text('description')``        | LONGTEXT equivalent column.                          |
+-------------------------------------------+------------------------------------------------------+
| ``table.medium_integer('votes')``         | MEDIUMINT equivalent column.                         |
+-------------------------------------------+------------------------------------------------------+
| ``table.medium_text('description')``      | MEDIUMTEXT equivalent column.                        |
+-------------------------------------------+------------------------------------------------------+
| ``table.small_integer('votes')``          | SMALLINT equivalent column.                          |
+-------------------------------------------+------------------------------------------------------+
| ``table.string('email')``                 | VARCHAR equivalent column.                           |
+-------------------------------------------+------------------------------------------------------+
| ``table.string('votes', 100)``            | VARCHAR equivalent column with a length.             |
+-------------------------------------------+------------------------------------------------------+
| ``table.text('description')``             | TEXT equivalent column.                              |
+-------------------------------------------+------------------------------------------------------+
| ``table.time('sunrise')``                 | TIME equivalent column.                              |
+-------------------------------------------+------------------------------------------------------+
| ``table.timestamp('added_at')``           | TIMESTAMP equivalent column.                         |
+-------------------------------------------+------------------------------------------------------+

Column Modifiers
----------------
Each of the column methods above returns a :py:class:`.Column` instance, which lets you add several
column "modifier" while adding a column to a table. For example, to make a column "nullable" you can
use the :py:meth:`~.Column.nullable` method::

    with self.schema.table('users') as table:
        table.string('email').nullable()

The column modifiers are the following *(does not include index modifiers)*:

=============================  ==============================================
Modifier                       Description
=============================  ==============================================
``.default('default_value')``  Specify a default value for the column.
``.nullable()``                Designate that the column allows NULL values.
``.unsigned()``                Set INTEGER column as UNSINGED *(MySQL)*.
=============================  ==============================================

Dropping Columns
----------------
To drop a column from a table, you can use the :py:meth:`~.Blueprint.drop_column`
method.

Dropping a column from a database table::

    with self.schema.table('users') as table:
        table.drop_column('votes')

Dropping multiple columns from a database table::

    with self.schema.table('users') as table:
        table.drop_column('votes', 'avatar', 'location')

Indexes
=======
The schema builder supports several types of indexes which you can add to and/or drop from columns.
You can create the index after defining the column. For example::

    table.string('email')
    table.unique('email')

Available Index Types
---------------------
Each index method accepts an optional second argument to specify the name of the index.
If omitted, the name will be derived from the names of the table and column(s).
Below is a list of all available index types:

======================================  =====================
Command                                 Description
======================================  =====================
``table.primary('id')``                 Addys a primary key.
``table.primary(['id', 'parent_id'])``  Adds composite keys.
``table.unique('email)``                Adds a unique index.
``table.index('state')``                Adds a basic index.
======================================  =====================

.. note::
    In ``MySQL``/``MariaDB`` and ``PostgreSQL`` the length of indexes are limited.
    experimentum generates index names based on table and column names, therefore if
    your column names are too long you can pass the ``name`` keyword argument to specify
    your own index name::

        table.index(
            ['some_field_with_a_really_long_name', 'another_really_long_field'],
            name='my_idx_name'
        )

Dropping Indexes
----------------
To drop an index you must specify the index's columns. experimentum assigns a reasonable
name to the indexes by default. If you have specified a custom index name, you have to add
it as the optional second argument.

=================================  =============================================
Command                            Description
=================================  =============================================
``table.drop_primary('id')``       Drop a primary key from the "id" column.
``table.drop_unique('email)``      Drop a unique index from the "email" column.
``table.drop_index('state')``      Drop a basic index from the "state" column.
``table.drop_foreign('user_id')``  Drop a foreign key from the "user_id" column.
=================================  =============================================


Foreign Key Contraints
======================
experimentum also provides support for adding foreign key constraints to your tables::

    with self.schema.table('posts') as table:
        table.increments('user_id')
        table.foreign('user_id')\
            .references('id').on('users')\
            .on_delete('cascade')\
            .on_update('cascade')

In this example, we are stating that the ``user_id`` column references the ``id`` column
on the ``users`` table and set the "on update" and "on delete" actions to ``cascade``.
Make sure to create the foreign key column first!

To drop a foreign key, you can use the :py:meth:`~.Blueprint.drop_foreign` method. It works
just like dropping an index.
"""
from experimentum.Storage.Migrations import Column, ForeignKey


class Blueprint(object):

    """Lets you create, delete, or alter columns of a table.

    Inspired by the Lavavel Schema Blueprint
    (https://laravel.com/docs/5.6/migrations#columns).

    Attributes:
        table (str): Name of the table.
        columns (list): Defaults to []. List of columns for the table.
        dropped (object): Defaults to {'columns': [], 'indexes': []}. Object with list of dropped
                        columns and indexes of the table.
        indexes (list): Defaults to []. List of indices for the table.
        fkeys (list): Defaults to []. List of foreign keys for the table.
        action (str): Defaults to 'alter'. Action (i.e. create or alter).
    """

    def __init__(self, table):
        """Initialize the blueprint.

        Args:
            table (str): Name of the table.
        """
        self.table = table
        self.columns = []
        self.dropped = {'columns': [], 'indexes': []}
        self.indexes = []
        self.fkeys = []
        self.action = 'alter'

    def create(self):
        """Set the create action."""
        self.action = 'create'

    ###############
    # * Columns * #
    ###############
    def add_column(self, col_type, name, **kwargs):
        """Add a new column to the blueprint.

        Args:
            col_type (str): Column Type
            name (str): Name of the Column.

        Returns:
            Column: Column data structure to add column modifiers.
        """
        col = Column(col_type, name, kwargs)
        self.columns.append(col)
        return col

    def drop_column(self, *args):
        """Drop one or multiple columns.

        Args:
            *args (str): Column names
        """
        self.dropped['columns'].extend(list(args))

    def array(self, column, arr_type, dimensions=None):
        """ARRAY column type.

        Only supported in **postgresql**, for other systems it defaults back to `arr_type`!

        Args:
            column (str): Name of the column.
            arr_type (str): Type of array
            dimensions (int, optional): Defaults to None. Array Dimension.

        Returns:
            Column: Column data structure to add column modifiers.
        """
        return self.add_column('array', column, arr_type=arr_type, dimensions=dimensions)

    def big_increments(self, column):
        """Auto-incrementing ID using a "big integer" (primary key) equivalent column.

        Args:
            column (str): Name of the column.

        Returns:
            Column: Column data structure to add column modifiers.
        """
        return self.add_column('big_integer', column, autoincrement=True)

    def big_integer(self, column):
        """BIGINT equivalent column.

        Args:
            column (str): Name of the column.

        Returns:
            Column: Column data structure to add column modifiers.
        """
        return self.add_column('big_integer', column)

    def binary(self, column):
        """BLOB equivalent column.

        Args:
            column (str): Name of the column.

        Returns:
            Column: Column data structure to add column modifiers.
        """
        return self.add_column('binary', column)

    def boolean(self, column):
        """BOOLEAN equivalent column.

        Args:
            column (str): Name of the column.

        Returns:
            Column: Column data structure to add column modifiers.
        """
        return self.add_column('boolean', column)

    def char(self, column, length):
        """CHAR equivalent column with a length.

        Args:
            column (str): Name of the column.
            length (int): Length of the column

        Returns:
            Column: Column data structure to add column modifiers.
        """
        return self.add_column('char', column, length=length)

    def date(self, column):
        """DATE equivalent column.

        Args:
            column (str): Name of the column.

        Returns:
            Column: Column data structure to add column modifiers.
        """
        return self.add_column('date', column)

    def datetime(self, column):
        """DATETIME equivalent column.

        Args:
            column (str): Name of the column.

        Returns:
            Column: Column data structure to add column modifiers.
        """
        return self.add_column('datetime', column)

    def decimal(self, column, precision, scale=2):
        """DECIMAL equivalent column with a precision (total digits) and a scale (decimal digits).

        Args:
            column (str): Name of the column.
            precision (int): Total Digits
            scale (int, optional): Defaults to 2. Number of decimal gigits

        Returns:
            Column: Column data structure to add column modifiers.
        """
        return self.add_column('decimal', column, precision=precision, scale=scale)

    def double(self, column, precision, scale=2):
        """DOUBLE equivalent column with a precision (total digits) and a scale (decimal digits).

        Args:
            column (str): Name of the column.
            precision (int): Total Digits
            scale (int): Default to 2. Number of decimal digits.

        Returns:
            Column: Column data structure to add column modifiers.
        """
        return self.add_column('double', column, precision=precision, scale=scale)

    def enum(self, column, fields):
        """ENUM equivalent column.

        Args:
            column (str): Name of the column.
            fields (list): Field Names for the enum

        Returns:
            Column: Column data structure to add column modifiers.
        """
        return self.add_column('enum', column, fields=fields)

    def float(self, column):
        """FLOAT equivalent column.

        Args:
            column (str): Name of the column.

        Returns:
            Column: Column data structure to add column modifiers.
        """
        return self.add_column('float', column)

    def increments(self, column):
        """Auto-incrementing ID using a "integer" (primary key) equivalent column.

        Args:
            column (str): Name of the column.

        Returns:
            Column: Column data structure to add column modifiers.
        """
        return self.add_column('integer', column, autoincrement=True)

    def integer(self, column):
        """INTEGER equivalent column.

        Args:
            column (str): Name of the column.

        Returns:
            Column: Column data structure to add column modifiers.
        """
        return self.add_column('integer', column)

    def json(self, column):
        """JSON equivalent column.

        Args:
            column (str): Name of the column.

        Returns:
            Column: Column data structure to add column modifiers.
        """
        return self.add_column('json', column)

    def long_text(self, column):
        """LONGTEXT equivalent column.

        Args:
            column (str): Name of the column.

        Returns:
            Column: Column data structure to add column modifiers.
        """
        return self.add_column('long_text', column)

    def medium_integer(self, column):
        """MEDIUMINT equivalent column.

        Args:
            column (str): Name of the column.

        Returns:
            Column: Column data structure to add column modifiers.
        """
        return self.add_column('medium_integer', column)

    def medium_text(self, column):
        """MEDIUMTEXT equivalent column.

        Args:
            column (str): Name of the column.

        Returns:
            Column: Column data structure to add column modifiers.
        """
        return self.add_column('medium_text', column)

    def small_integer(self, column):
        """SMALLINT equivalent column.

        Args:
            column (str): Name of the column.

        Returns:
            Column: Column data structure to add column modifiers.
        """
        return self.add_column('small_integer', column)

    def string(self, column, length=None):
        """VARCHAR equivalent column with an optional length.

        Args:
            column (str): Name of the column.
            length (int, optional): Defaults to None. Length of the string.

        Returns:
            Column: Column data structure to add column modifiers.
        """
        return self.add_column('string', column, length=length)

    def text(self, column):
        """TEXT equivalent column.

        Args:
            column (str): Name of the column.

        Returns:
            Column: Column data structure to add column modifiers.
        """
        return self.add_column('text', column)

    def time(self, column):
        """TIME equivalent column.

        Args:
            column (str): Name of the column.

        Returns:
            Column: Column data structure to add column modifiers.
        """
        return self.add_column('time', column)

    def timestamp(self, column):
        """TIMESTAMP equivalent column.

        Args:
            column (str): Name of the column.

        Returns:
            Column: Column data structure to add column modifiers.
        """
        return self.add_column('timestamp', column)

    ###############
    # * Indexes * #
    ###############
    def primary(self, column, name=None):
        """Add a primary key to the column.

        Args:
            column (str): Primary Key column
            name (str, optional): Default to None. Name of the primary key.

        Returns:
            Blueprint: self instance for method chaining.
        """
        self._add_idx(column, 'primary', name)
        return self

    def unique(self, column, name=None):
        """Add a unique key to the column.

        Args:
            column (str): Unique Key column
            name (str, optional): Default to None. Name of the unique key.

        Returns:
            Blueprint: self instance for method chaining.
        """
        self._add_idx(column, 'unique', name)
        return self

    def index(self, column, name=None):
        """Add a basic index to the column.

        Args:
            column (str): Indexed column
            name (str, optional): Default to None. Name of the unique key.

        Returns:
            Blueprint: self instance for method chaining.
        """
        self._add_idx(column, 'index', name)
        return self

    def foreign(self, column, name=None):
        """Add a foreign key to the column.

        Args:
            column (str): Foreign Key column
            name (str, optional): Default to None. Name of the foreign key.

        Returns:
            ForeignKey: ForeignKey data structure to configure foreign key.
        """
        if not name:
            name = '{}_{}_{}'.format(self.table, column, 'foreign')

        fkey = ForeignKey(column, name)
        self.fkeys.append({'col': column, 'key': fkey})
        return fkey

    def _add_idx(self, column, idx_type, name=None):
        """Add an index to the index list.

        Args:
            column (str): indexed column.
            idx_type (str): Type of the index.
            name (str, optional): Default to None. Custom index name.
        """
        if not name:
            name = '{}_{}_{}'.format(self.table, column, idx_type)

        self.indexes.append({
            'col': column,
            'type': idx_type,
            'name': name
        })

    def drop_primary(self, column, name=None):
        """Drop a primary key from the column.

        Args:
            column (str): Primary Key column
            name (str, optional): Default to None. Name of the primary key.
        """
        self._drop_idx(column, 'primary', name)

    def drop_unique(self, column, name=None):
        """Drop a unique key from the column.

        Args:
            column (str): unique Key column
            name (str, optional): Default to None. Name of the unique key.
        """
        self._drop_idx(column, 'unique', name)

    def drop_index(self, column, name=None):
        """Drop a index key from the column.

        Args:
            column (str): index Key column
            name (str, optional): Default to None. Name of the index key.
        """
        self._drop_idx(column, 'index', name)

    def drop_foreign(self, column, name=None):
        """Drop a foreign key from the column.

        Args:
            column (str): foreign Key column
            name (str, optional): Default to None. Name of the foreign key.
        """
        self._drop_idx(column, 'foreign', name)

    def _drop_idx(self, column, idx_type, name=None):
        """Drop an index.

        Args:
            column (str): indexed column
            idx_type (str): Type of the index.
            name (str, optional): Default to None. Custom index name.
        """
        if not name:
            name = '{}_{}_{}'.format(self.table, column, idx_type)

        self.dropped['indexes'].append({
            'col': column,
            'type': idx_type,
            'name': name
        })



"""Main Service Container and Provider.

Sets up the framework and runs the experiments and lets you customize/extend
the behavior of the framework.

Binding
-------
We can register/bind a new  alias by extending either the :py:meth:`~.App.register_aliases class`
or by directly adding the alias to the :py:attr:`~.App.aliases` dictionary. The key is the alias
name you want to register and the value is a function that returns an instance of the class::

    def register_aliases(self):
        super(MyAppClass, self).register_aliases()

        self.aliases['my_custom_api'] = lambda: API(self.store)

Additional arguments for creating a class instance may be passed when resolving. Your function
just has to add them in order to use them::

    self.aliases['my_custom_api'] = lambda name, user_id=None: API(self.store, name, user_id)

Resolving
---------
You may use the :py:meth:`~.App.make` method to resolve a class instance out of the container.
The :py:meth:`~.App.make` method accepts the alias of the class you want to resolve::

    api = self.app.make('my_custom_api')

If some of your class' dependencies are not resolvable via the app container, you may pass them
as additional args and keyword args::

    api = self.app.make('my_custom_api', foo, user_id=42)

Customizing
-----------

Commands
^^^^^^^^
Register new commands with the :py:meth:`.App.register_commands` method.
It should return a dictionary where the keys are names of the commands and
the values are the command handlers. The command handlers must either be
derived from :py:class:`.AbstractCommand` or a function with the decorator
:py:func:`.AbstractCommand.command`. Example return::

    {
        'foo': FooCommand,
        'bar': BarCommand
    }
"""
from __future__ import print_function
import os
import sys
import logging
import logging.handlers
from collections import OrderedDict
from sqlalchemy import create_engine
from sqlalchemy.engine.url import URL
from experimentum.cli import print_failure
from experimentum.Config import Config, Loader
from experimentum.Commands import CommandManager, MigrationCommand, ExperimentsCommand,\
    PlotCommand, WebGUICommand
from experimentum.Experiments import Experiment
from experimentum.Storage.AbstractStore import AbstractStore
from experimentum.Storage.AbstractRepository import RepositoryLoader
from experimentum.Storage.Migrations import Migrator, Blueprint, Schema
from experimentum.Storage.SQLAlchemy import Store, Repository
from experimentum.Plots import Factory
from experimentum.WebGUI import Server


def _path_join(root, path):
    """Join root with relative path.

    Args:
        root (str): Root path
        path (str): relative path to root

    Returns:
        str: realpath
    """
    return os.path.realpath(os.path.join(root, path))


class App(object):

    """Main entry point of the framework.

    Attributes:
        config_path (str): Defaults to ``.``. Path to config files.
        base_repository (AbstractRepository): Defaults to ``Repository``. Repository Base Class
        name (str): Name of the app.
        root (str): Root Path of the app.
        config (Config): Config Manager.
        log (logging.Logger): Logger.
        store (AbstractStore): Data Store.
        aliases (dict): Dictionary of aliases and factory functions.
    """
    config_path = '.'
    base_repository = Repository

    def __init__(self, name, root):
        """Bootstrap the app framework.

        Args:
            name (str): Name of the App.
        """
        self.name = name
        self.root = os.path.dirname(root)
        self.log = logging.getLogger('experimentum')
        self.config = Config()
        self.store = None
        self.aliases = {}
        self.bootstrap()

    def bootstrap(self):
        """Bootstrap the app, i.e. setup config and logger."""
        # Load Config
        loader = Loader(_path_join(self.root, self.config_path), self.config)
        loader.load_config_files()

        # Add experiments and repo folders to path
        paths = [
            self.config.get('app.experiments.path', 'experiments'),
            self.config.get('storage.repositories.path', 'repositories')
        ]
        for path in paths:
            sys.path.insert(0, os.path.abspath(_path_join(self.root, path)))

        # Setup logger
        self._set_logger()
        self.log.info('Bootstrap App')

        # Setup Datastore
        self.log.info('Setup Data Store')
        default_db = {'drivername': 'sqlite', 'database': 'experimentum.db'}
        database = self.config.get('storage.datastore', default_db)

        # Absolute file path based on root for sqlite databases (except in-memory)
        if database['drivername'] == 'sqlite' and database['database'] != '':
            database['database'] = _path_join(self.root, database['database'])

        self.setup_datastore(database)

        if not isinstance(self.store, AbstractStore):
            msg = 'The "{}" Store implementation must implement the AbstractStore interface'
            print_failure(msg.format(self.store.__class__), 1)

        # Load and map Repositories
        self.repositories = RepositoryLoader(self, self.base_repository, self.store)
        self.repositories.load()

        # Aliases
        self.register_aliases()

        # Register commands
        self.log.info('Register Commands')
        self.cmd_manager = CommandManager(
            self,
            self.config.get('app.prog', 'app'),
            self.config.get('app.description', '')
        )
        self._add_commands()

        # save app instance
        # app(self)

    def setup_datastore(self, datastore):
        """Set up the data store.

        Args:
            datastore (dict): Datastore config.
        """
        self.store = Store(self)

        db_args = {}
        if datastore['drivername'] == 'sqlite':
            db_args['connect_args'] = {'check_same_thread': False}

        self.store.set_engine(
            create_engine(URL(**datastore), **db_args)
        )

    def make(self, alias, *args, **kwargs):
        """Create an instance of an aliased class.

        Args:
            alias (str): Name of class alias.

        Raises:
            Exception: if the alias does not exists.

        Returns:
            object: Instance of the aliased class
        """
        klass = self.aliases.get(alias, None)

        if klass is None:
            print_failure("Class with alias '{}' does not exist.".format(alias), 1)

        return klass(*args, **kwargs)

    def run(self):
        """Run the app."""
        self.cmd_manager.dispatch()
        self.log.info('Run App')

    def register_commands(self):
        """Register custom cli commands.

        Returns:
            dict: commands to register
        """
        return {}

    def register_aliases(self):
        """Register aliases for classes."""
        self.log.info('Register Aliases')

        migration_path = self.config.get('storage.migrations.path', 'migrations')
        experiments_path = self.config.get('app.experiments.path', 'experiments')
        plot_factory = Factory(self)

        self.aliases = {
            'experiment':
                lambda name: Experiment.load(self, _path_join(self.root, experiments_path), name),
            'migrator':
                lambda: Migrator(_path_join(self.root, migration_path), self),
            'plot': plot_factory.create,
            'store': lambda: self.store,
            'schema': lambda: Schema(self),
            'blueprint': Blueprint,
            'server': lambda: Server(self),
            'config': Config
        }

    def _add_commands(self):
        """Add Commands to the command manager."""
        commands = self._commands()
        user_commands = self.register_commands()

        if any(user_commands):
            commands.update(OrderedDict(sorted(user_commands.items(), key=lambda c: c[0])))

        for name, cmd in commands.items():
            self.log.debug('Adding Command {}'.format(name))
            self.cmd_manager.add_command(name, cmd)

    @classmethod
    def _commands(cls):
        """Register default commands.

        Returns:
            dict: Default commands
        """
        commands = OrderedDict()
        commands['experiments:run'] = ExperimentsCommand.run
        commands['experiments:list'] = ExperimentsCommand.status
        commands['migration:status'] = MigrationCommand.status
        commands['migration:refresh'] = MigrationCommand.refresh
        commands['migration:up'] = MigrationCommand.up
        commands['migration:down'] = MigrationCommand.down
        commands['migration:make'] = MigrationCommand.make
        commands['plot:generate'] = PlotCommand.generate
        commands['webgui'] = WebGUICommand.start

        return commands

    def _set_logger(self):
        """Set up the logger and its handlers."""
        # Set log level
        levels = {
            'debug': logging.DEBUG,
            'info': logging.INFO,
            'warning': logging.WARNING,
            'error': logging.ERROR,
            'critical': logging.CRITICAL
        }
        level = levels.get(
            self.config.get('app.logging.level', 'info'), logging.NOTSET
        )
        self.log.setLevel(level)

        # Add filehandler and Formatter
        name = self.config.get(
            'app.logging.filename',
            '{name}.log'
        ).format(name=self.name).lower()

        filename = '{}/{}'.format(
            os.path.realpath(_path_join(self.root, self.config.get('app.logging.path', '.'))),
            name
        )
        filehandler = logging.handlers.RotatingFileHandler(
            filename,
            maxBytes=self.config.get('app.logging.max_bytes', 1024 * 1024),  # default to 1MB
            backupCount=self.config.get('app.logging.backup_count', 10),
            mode='a+'
        )
        filehandler.setLevel(level)
        filehandler.setFormatter(logging.Formatter(self.config.get(
            'app.logging.format',
            '[%(asctime)s] - [%(levelname)s] - [%(module)s] - %(message)s'
        )))
        self.log.addHandler(filehandler)
